import math

import numpy as np
from numba import jit, njit, prange
import copy
import scipy.signal
import matplotlib.pyplot as plt
import torch


def widen_boolean(arr, n_before, n_after, axis=None):
    '''
    Widens boolean events by n_before and n_after.    
    RH 2021    

    Args:
        arr (np.ndarray):
            Input array. Widening will be applied
             to the last dimension.
        n_before (int):
            Number of samples before 'True' values
             that will also be set to 'True'.
        n_after (int):
            Number of samples after 'True' values
             that will also be set to 'True'.
        axis (int):
            Axis to apply the event widening.
             If None then arr should be a 1-D array.
    
    Returns:
        widened arr (np.ndarray, dtype=bool):
            Output array. Same as input arr, but
             with additional 'True' values before
             and after initial 'True' values.
    '''
    
    kernel = np.zeros(np.max(np.array([n_before, n_after])) * 2 + 1)
    kernel_center = int(np.ceil(len(kernel) / 2))
    kernel[kernel_center - (n_before+1): kernel_center] = 1
    kernel[kernel_center: kernel_center + n_after] = 1
    kernel = kernel / np.mean(kernel)
    
    if axis is None:
        return np.bool_(scipy.signal.convolve(arr, kernel, mode='same'))
    else:
        return np.bool_(np.apply_along_axis(lambda m: scipy.signal.convolve(m, kernel, mode='same'),
                                                              axis=axis, arr=arr))


# @njit
def idx2bool(idx, length=None):
    '''
    Converts a vector of indices to a boolean vector.
    RH 2021

    Args:
        idx (np.ndarray):
            1-D array of indices.
        length (int):
            Length of boolean vector.
            If None then length will be set to
             the maximum index in idx + 1.
    
    Returns:
        bool_vec (np.ndarray):
            1-D boolean array.
    '''
    ## remove NaNs
    idx = idx[~np.isnan(idx)].astype(np.int64)
    if length is None:
        length = np.uint64(np.max(idx) + 1)
    out = np.zeros(length, dtype=np.bool8)
    out[idx] = True
    return out


def bool2idx(bool_vec):
    '''
    Converts a boolean vector to indices.
    RH 2021

    Args:
        bool_vec (np.ndarray):
            1-D boolean array.
    
    Returns:
        idx (np.ndarray):
            1-D array of indices.
    '''
    return np.where(bool_vec)[0]


@njit
def binary_search(arr, lb, ub, val):
    '''
    Recursive binary search
    adapted from https://www.geeksforgeeks.org/python-program-for-binary-search/
    RH 2021
    
    Args:
        arr (sorted list):
            1-D array of numbers that are already sorted.
            To use numba's jit features, arr MUST be a
            typed list. These include:
                - numpy.ndarray (ideally np.ascontiguousarray)
                - numba.typed.List
        lb (int):
            lower bound index.
        ub (int):
            upper bound index.
        val (scalar):
            value being searched for
    
    Returns:
        output (int):
            index of val in arr.
            returns -1 if value is not present
            
    Demo:
        # Test array
        arr = np.array([ 2, 3, 4, 10, 40 ])
        x = 100

        # Function call
        result = binary_search(arr, 0, len(arr)-1, x)

        if result != -1:
            print("Element is present at index", str(result))
        else:
            print("Element is not present in array")
    '''
    # Check base case
    if ub >= lb:
 
        mid = (ub + lb) // 2
 
        # If element is present at the middle itself
        if arr[mid] == val:
            return mid
 
        # If element is smaller than mid, then it can only
        # be present in left subarray
        elif arr[mid] > val:
            return binary_search(arr, lb, mid - 1, val)
 
        # Else the element can only be present in right subarray
        else:
            return binary_search(arr, mid + 1, ub, val)
 
    else:
        # Element is not present in the array
        return -1


def get_last_True_idx(input_array):
    '''
    for 1-d arrays only. gets idx of last entry
    that == True
    RH 2021
    '''
    nz = np.nonzero(input_array)[0]
    # print(nz.size)
    if nz.size==0:
        raise ValueError('No True values in array')
    else:
        output = np.max(nz)
#     print(output)
    return output

def get_nth_True_idx(input_array, n):
    '''
    for 1-d arrays only. gets idx of nth True entry.
    nth is zero-indexed.
    RH 2022
    '''
    nz = np.nonzero(input_array)[0]
    # print(nz.size)
    if nz.size==0:
        raise ValueError('No True values in array')
    else:
        output = nz[n]
    return output


def make_batches(
    iterable, 
    batch_size=None, 
    num_batches=None, 
    min_batch_size=0, 
    return_idx=False, 
    length=None,
    idx_start=0,
    randomize_batch_indices=False,
    shuffle_batch_order=False,
    shuffle_iterable_order=False,
):
    """
    Make batches of data or any other iterable.
    RH 2021-2024

    Args:
        iterable (iterable):
            Iterable to be batched.
        batch_size (int):
            Size of each batch.\n
            if None, then ``batch_size`` based on ``num_batches``.
        num_batches (int):
            Number of batches to make.\n
            if None, then ``num_batches`` based on ``batch_size``.
        min_batch_size (int):
            Minimum size of each batch. Set to ``-1`` to equal ``batch_size``.
        return_idx (bool):
            Whether to also return the slice indices of the batches. Returns will
            be the tuple: ``(iterable[idx], [idx_start, idxend])``
        length (int):
            Length of the iterable. Determines ``idx_end``. If ``None``, then
            length is ``len(iterable)`` This is useful if you want to make
            batches of something that doesn't have a ``__len__`` method.
        idx_start (int):
            Starting index of the iterable.
        randomize_batch_indices
            Whether to randomize the transition indices between batches.\n
            So ``[(1,2,3), (4,5,6),]`` can become ``[(1,), (2,3,4), (5,6),]``.
        shuffle_batch_order (bool):
            Whether to shuffle the order in which batches are yielded.
        shuffle_iterable_order (bool):
            Whether to shuffle the order of the contents of the iterable before
            batching.
    
    Returns:
        output (iterable):
            batches of iterable
    """

    ## Prepare parameters
    l = len(iterable) if length is None else length

    assert sum([(batch_size is None), (num_batches is None)]) == 1, 'Exactly one of batch_size or num_batches must be specified. The other must be None.'
    if batch_size is None:
        assert isinstance(num_batches, int), 'num_batches must be an integer'
        batch_size = int(math.ceil(l / num_batches))
    if num_batches is None:
        assert isinstance(batch_size, int), 'batch_size must be an integer'
        num_batches = int(math.ceil(l / batch_size))

    if min_batch_size == -1:
        min_batch_size = batch_size

    assert batch_size > 0, 'batch_size must be greater than 0'
    assert min_batch_size <= batch_size, 'min_batch_size must be less than or equal to batch_size'
    assert num_batches > 0, 'num_batches must be greater than 0'    

    ## Make slices
    offset = np.random.randint(0, batch_size) if randomize_batch_indices else 0
    idx_slices = [slice(s, min(e, l), None) for s, e in (zip(np.arange(idx_start + offset, l, batch_size), np.arange(idx_start + batch_size + offset, l + batch_size + offset, batch_size)))]
    if offset > 0:
        idx_slices = [slice(0, offset, None)] + idx_slices

    ## Remove small batches
    if min_batch_size > 0:
        idx_slices = [idx for idx in idx_slices if (idx.stop - idx.start) >= min_batch_size]

    ## Shuffle batch order
    if shuffle_batch_order:
        idx_slices = [idx_slices[ii] for ii in np.random.permutation(len(idx_slices))]

    ## Crop to num_batches
    idx_slices = idx_slices[:num_batches]

    ## Shuffle iterable contents order
    if shuffle_iterable_order:
        order = np.random.permutation(l)
    else:
        order = range(l)

    ## Yield batches
    if return_idx:
        for idx in idx_slices:
            yield iterable[order[idx]], [order[idx.start], order[idx.stop]]
    else:
        for idx in idx_slices:
            yield iterable[order[idx]]


@njit
def find_nearest_idx(array, value):
    '''
    Finds the value and index of the nearest
     value in an array.
    RH 2021, 2024
    
    Args:
        array (np.ndarray):
            Array of values to search through.
        value (scalar):
            Value to search for.

    Returns:
        array_idx (int):
            Index of the nearest value in array.
    '''
    idx = np.searchsorted(array, value, side="left")
    if idx > 0 and (idx == len(array) or np.abs(value - array[idx-1]) < np.abs(value - array[idx])):
        return idx-1
    else:
        return idx
@njit(parallel=True)
def find_nearest_array(array, values, max_diff=None):
    '''
    Finds the values and indices of the nearest
     values in an array.
    RH 2021, 2024

    Args:
        array (np.ndarray):
            Array of values to search through.
        values (np.ndarray):
            Values to search for.

    Returns:
        array_idx (np.ndarray):
            Indices of the nearest values in array.
        array_val (np.ndarray):
            Values of the nearest values in array.
        diff (np.ndarray):
            Differences between the values and the
             nearest values in array.
    '''
    assert array.ndim == 1, 'array must be 1-D'
    assert values.ndim == 1, 'values must be 1-D'
    
    vals_nearest = np.zeros(values.shape if array.size > 0 else (0,), dtype=array.dtype)
    idx_nearest  = np.zeros(values.shape if array.size > 0 else (0,), dtype=np.int64)
    diff_nearest = np.zeros(values.shape if array.size > 0 else (0,), dtype=array.dtype)
    
    if array.size > 0:
        for ii in prange(len(values)):
            idx_nearest[ii] = find_nearest_idx(array , values[ii])

        vals_nearest = array[idx_nearest]
        diff_nearest = np.abs(vals_nearest - values)

        if max_diff is not None:
            bool_keep = diff_nearest <= max_diff
            vals_nearest = vals_nearest[bool_keep]
            idx_nearest = idx_nearest[bool_keep]
            diff_nearest = diff_nearest[bool_keep]
            
    return vals_nearest, idx_nearest, diff_nearest


def pad_with_singleton_dims(array, n_dims_pre=0, n_dims_post=0):
    arr_out = copy.copy(np.array(array))
    
    for n in range(n_dims_pre):
        arr_out = np.expand_dims(arr_out, 0)
    for n in range(n_dims_post):
        arr_out = np.expand_dims(arr_out, -1)
    return arr_out


def index_with_nans(values, indices):
    """
    Indexes an array with a list of indices, allowing for NaNs in the indices.
    RH 2022
    
    Args:
        values (np.ndarray):
            Array to be indexed.
        indices (Union[List[int], np.ndarray]):
            1D list or array of indices to use for indexing. Can contain NaNs.
            Datatype should be floating point. NaNs will be removed and values
            will be cast to int.

    Returns:
        np.ndarray:
            Indexed array. Positions where `indices` was NaN will be filled with
            NaNs.
    """
    indices = np.array(indices, dtype=float) if not isinstance(indices, np.ndarray) else indices
    values = np.concatenate((np.full(shape=values.shape[1:], fill_value=np.nan, dtype=values.dtype)[None,...], values), axis=0)
    idx = indices.copy() + 1
    idx[np.isnan(idx)] = 0
    
    return values[idx.astype(np.int64)]


def shift_pad(array, shift=1, axis=-1, pad_val=0, in_place=False):
    """
    Pads an array with a constant value.
    Allows for shifting along any axis.
    RH 2022

    Args:
        array (np.ndarray):
            array to shift and pad
        shift (int):
            number of elements to shift
        axis (int):
            axis to shift along
        pad_val (any):
            value to pad with
        in_place (bool):
            whether to shift and pad in place

    Returns:
        output (np.ndarray):
            shifted and padded array
    """
    if shift==0:
        return array
        
    if shift > 0:
        idx = np.arange(array.shape[axis])[:-shift]
    if shift < 0:
        idx = np.arange(array.shape[axis])[-shift:]
    
    if axis==-1:
        axis_to_nix = array.shape[-1]
    else:
        axis_to_nix = axis
                
    dims_to_append = list(array.shape)
    dims_to_append[axis_to_nix] = np.abs(shift)
    
    padding = np.ones(dims_to_append) * pad_val
    
    if in_place:
        out = array
    else:
        out = None
    
    if shift > 0:
        arr_shifted = np.concatenate(
            ( padding, np.take(array, idx, axis=axis)),
            axis=axis,
            out=out
        )
    if shift < 0:
        arr_shifted = np.concatenate(
            ( np.take(array, idx, axis=axis),   padding ),
            axis=axis,
            out=out
        )
        
    return arr_shifted
    

def off_diagonal(x):
    """
    Returns the off-diagonal elements of a matrix as a vector.
    RH 2022

    Args:
        x (np.ndarray or torch tensor):
            square matrix to extract off-diagonal elements from.

    Returns:
        output (np.ndarray or torch tensor):
            off-diagonal elements of x.
    """
    n, m = x.shape
    assert n == m
    return x.reshape(-1)[:-1].reshape(n - 1, n + 1)[:, 1:].reshape(-1)


#######################################
############ SPARSE STUFF #############
#######################################

def sparse_mask(x, mask_sparse, do_safety_steps=True):
    """
    Masks a sparse matrix with the non-zero elements of another
     sparse matrix.
    RH 2022

    Args:
        x (scipy.sparse.csr_matrix):
            sparse matrix to mask
        mask_sparse (scipy.sparse.csr_matrix):
            sparse matrix to mask with
        do_safety_steps (bool):
            whether to do safety steps to ensure that things
             are working as expected.

    Returns:
        output (scipy.sparse.csr_matrix):
            masked sparse matrix
    """
    if do_safety_steps:
        m = mask_sparse.copy()
        m.eliminate_zeros()
    else:
        m = mask_sparse
    return (m!=0).multiply(x)

def scipy_sparse_to_torch_coo(sp_array):
    import torch

    coo = scipy.sparse.coo_matrix(sp_array)
    
    values = coo.data
    indices = np.vstack((coo.row, coo.col))

    i = torch.LongTensor(indices)
    v = torch.FloatTensor(values)
    shape = coo.shape

    return torch.sparse_coo_tensor(i, v, torch.Size(shape))

def pydata_sparse_to_torch_coo(sp_array):
    import sparse
    import torch

    coo = sparse.COO(sp_array)
    
    values = coo.data
#     indices = np.vstack((coo.row, coo.col))
    indices = coo.coords

    i = torch.LongTensor(indices)
    v = torch.FloatTensor(values)
    shape = coo.shape

    return torch.sparse_coo_tensor(i, v, torch.Size(shape))
    
def pydata_sparse_to_spconv(sp_array):
    import sparse
    import torch
    import spconv

    coo = sparse.COO(sp_array)
    idx_raw = torch.as_tensor(coo.coords.T, dtype=torch.int32)
    idx = torch.hstack((torch.zeros((idx_raw.shape[0],1)) , idx_raw)).type(torch.int32)
    spconv_array = rois_sp_spconv = spconv.SparseConvTensor(
        features=coo.reshape((coo.shape[0], -1)).T,
        indices=idx,
        spatial_shape=coo.shape, 
        batch_size=1
    )
    return spconv_array

def sparse_convert_spconv_to_scipy(sp_arr):
    import sparse

    coo = sparse.COO(
        coords=sp_arr.indices.T.to('cpu'),
        data=sp_arr.features.squeeze().to('cpu'),
        shape=[sp_arr.batch_size] + sp_arr.spatial_shape
    )
    return coo.reshape((coo.shape[0], -1)).to_scipy_sparse().tocsr()

def torch_to_torchSparse(s):
    import torch_sparse

    # return torch_sparse.from_torch_sparse(s)

    return torch_sparse.tensor.SparseTensor(
        row=s.indices()[0],
        col=s.indices()[1],
        value=s.values(),
        sparse_sizes=s.shape,
    )

# def pydata_sparse_to_torchSparse(s, shape=None):
#     import torch
#     import sparse
#     import torch_sparse

#     return torch_sparse.from_torch_sparse(pydata_sparse_to_torch_coo(s).coalesce())

    # coords = [torch.LongTensor(c) for c in s.coords]
    # vals = torch.as_tensor(s)

    # return torch_sparse.tensor.SparseTensor(
    #     row=coords[0],
    #     col=coords[1],
    #     value=vals,
    #     sparse_sizes=shape,
    # )

def denseDistances_to_knnDistances(denseDistanceMatrix, k=1023, epsilon=1e-9):
    """
    Converts a dense distance matrix to a sparse kNN distance matrix.
    Largest values are sparsened away. Zeros are set to epsilon.
    Useful for converting custom distance matrices into a format that
     can be used by things like sklearn's nearest neighbors algorithms.
    RH 2022

    Args:
        denseDistanceMatrix (np.ndarray):
            Dense distance matrix
        k (int):
            Number of nearest neighbors to find
        epsilon (float):
            Small number to add to distances because entries with
             distance 0 are ignored by scipy's sparse csr_matrix.
            It can be subtracted out later.

    Returns:
        output (scipy.sparse.csr_matrix):
            Sparse kNN distance matrix
    """
    X = denseDistanceMatrix + epsilon
    k_lowest = np.argsort(X, axis=1)[:,:k]
    kl_bool = np.stack([idx2bool(k_l, length=X.shape[1]) for k_l in k_lowest])
    X[~kl_bool] = 0
    return scipy.sparse.csr_matrix(X), kl_bool


class scipy_sparse_csr_with_length(scipy.sparse.csr_matrix):
    """
    A scipy sparse matrix with a length attribute.
    Useful when needing to iterate over the rows of a sparse matrix.
    This class gives a length attribute to the sparse matrix.
    RH 2022
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.length = self.shape[0]

    def __len__(self):
        return self.length

    def __getitem__(self, key):
        return self.__class__(super().__getitem__(key))


def find_nonredundant_idx(s):
    """
    Finds the indices of the nonredundant entries in a sparse matrix.
    Useful when you are manually populating a spare matrix and want to
     know which entries you have already populated.
    RH 2022

    Args:
        s (scipy.sparse.coo_matrix):
            Sparse matrix. Should be in coo format.

    Returns:
        idx_unique (np.ndarray):
            Indices of the nonredundant entries
    """
    if s.getformat() != 'coo':
        s = s.coo()
    idx_rowCol = np.vstack((s.row, s.col)).T
    u, idx_u = np.unique(idx_rowCol, axis=0, return_index=True)
    return idx_u
def remove_redundant_elements(s, inPlace=False):
    """
    Removes redundant entries from a sparse matrix.
    Useful when you are manually populating a spare matrix and want to
     remove redundant entries.
    RH 2022

    Args:
        s (scipy.sparse.coo_matrix):
            Sparse matrix. Should be in coo format.
        inPlace (bool):
            If True, the input matrix is modified in place.
            If False, a new matrix is returned.

    Returns:
        s (scipy.sparse.coo_matrix):
            Sparse matrix with redundant entries removed.
    """
    idx_nonRed = find_nonredundant_idx(s)
    if inPlace:
        s_out = s
    else:
        s_out = scipy.sparse.coo_matrix(s.shape)
    s_out.row = s.row[idx_nonRed]
    s_out.col = s.col[idx_nonRed]
    s_out.data = s.data[idx_nonRed]
    return s_out

def merge_sparse_arrays(s_list, idx_list, shape_full, remove_redundant=True, elim_zeros=True):
    """
    Merges a list of square sparse arrays into a single square sparse array.
    Note that no selection is performed for removing redundant entries;
     just whatever is selected by np.unique is kept.

    Args:
        s_list (list of scipy.sparse.csr_matrix):
            List of sparse arrays to merge.
            Each array can be of any shape.
        idx_list (list of np.ndarray int):
            List of arrays of integers. Each array should be of the same
             length as the corresponding array in s_list and should contain
             integers in the range [0, shape_full[0]). These integers
             represent the row/col indices in the full array.
        shape_full (tuple of int):
            Shape of the full array.
        remove_redundant (bool):
            If True, redundant entries are removed from the output array.
            If False, redundant entries are kept.

    Returns:
        s_full (scipy.sparse.csr_matrix):
    """
    row, col, data = np.array([]), np.array([]), np.array([])
    for s, idx in zip(s_list, idx_list):
        s_i = s.tocsr() if s.getformat() != 'csr' else s
        s_i.eliminate_zeros() if elim_zeros else s_i
        idx_grid = np.meshgrid(idx, idx)
        row = np.concatenate([row, (s_i != 0).multiply(idx_grid[0]).data])
        col = np.concatenate([col, (s_i != 0).multiply(idx_grid[1]).data])
        data = np.concatenate([data, s_i.data])
    s_full = scipy.sparse.coo_matrix((data, (row, col)), shape=shape_full)
    if remove_redundant:
        remove_redundant_elements(s_full, inPlace=True)
    return s_full


def sparse_to_dense_fill(arr_s, fill_val=0.):
    """
    Converts a sparse array to a dense array and fills
     in sparse entries with a fill value.
    """
    import sparse
    s = sparse.COO(arr_s)
    s.fill_value = fill_val
    return s.todense()


####################################################################################################
##################################### TENSOR OPERATIONS ############################################
####################################################################################################

def cp_to_kruskal(cp):
    """
    Converts a list (of length n_modes) of 2D arrays (of shape (len_dim, rank))
     [CP format] to a list (of length rank) of lists (of length n_modes) of
     vectors (of length len_dim) [Kruskal format].
    RH 2022

    Args:
        cp (list of np.ndarray):
            List of 2D arrays in CP format.
            Tensorly uses this format for their 'cp' format.

    Returns:
        k (list of list of np.ndarray):
            List of lists of vectors in Kruskal format.
    """
    rank = cp[0].shape[1]
    k = [[cp[m][:,r] for m in range(len(cp))] for r in range(rank)]
    return k

def kruskal_to_dense(k, weights=None):
    """
    Converts a list (of length rank) of lists (of length n_modes) of
     vectors (of length len_dim) [Kruskal format] to a dense tensor
     (of shape (len_dim, len_dim, ...))
    RH 2022

    Args:
        k (list of list of np.ndarray):
            List of lists of vectors in Kruskal format.

    Returns:
        dense (np.ndarray):
            Dense tensor
    """
    assert isinstance(k, list), 'k must be a list'
    assert isinstance(k[0], list), 'k must be a list of lists'
    assert all([len(k[0]) == len(k[r]) for r in range(len(k))]), 'each mode must have the same size'


    ## check if numpy or torch
    if isinstance(k[0][0], np.ndarray):
        zeros = np.zeros
        einsum = np.einsum
        weights = np.array(weights).astype(k[0][0].dtype) if weights is not None else np.ones(len(k)).astype(k[0][0].dtype)
    elif isinstance(k[0][0], torch.Tensor):
        zeros = torch.zeros
        einsum = torch.einsum
        weights = torch.as_tensor(weights).type(k[0][0].dtype).to(k[0][0].device) if weights is not None else torch.ones(len(k)).type(k[0][0].dtype).to(k[0][0].device)

    rank = len(k)
    n_modes = len(k[0])
    dense = zeros([k[0][m].shape[0] for m in range(n_modes)]).type(k[0][0].dtype).to(k[0][0].device)
    for r in range(rank):
        ## take the outer product of the n vectors in each mode and add to the dense tensor
        str_einsum = ','.join([chr(97+m) for m in range(n_modes)]) + '->' + ''.join([chr(97+m) for m in range(n_modes)])
        dense += einsum(str_einsum, *k[r]) * weights[r]

    return dense

def cp_to_dense(cp, weights=None):
    """
    Converts a list (of length n_modes) of 2D arrays (of shape (len_dim, rank))
     [CP format] to a dense tensor (of shape (len_dim, len_dim, ...))
    RH 2022

    Args:
        cp (list of np.ndarray):
            List of 2D arrays in CP format.
            Tensorly uses this format for their 'cp' format.

    Returns:
        dense (np.ndarray):
            Dense tensor
    """
    rank = cp[0].shape[1]
    n_modes = len(cp)
    str_einsum = ','.join([chr(97+m)+'r' for m in range(n_modes)]) + '->' + ''.join([chr(97+m) for m in range(n_modes)])
    if weights is None:
        weights = np.ones(rank)

    ## check if numpy or torch
    if isinstance(cp[0], np.ndarray):
        einsum = np.einsum
        weights = np.array(weights).astype(cp[0].dtype)
    elif isinstance(cp[0], torch.Tensor):
        einsum = torch.einsum
        weights = torch.as_tensor(weights).type(cp[0].dtype).to(cp[0].device)

    dense = einsum(str_einsum, *[cp[m] * weights for m in range(n_modes)])
    
    return dense


import copy

import torch
import numpy as np
from tqdm import tqdm

class RegressionRNN(torch.nn.Module):
    """
    RNN for time series-like regression

    RH 2023

    Args:
        input_size (int):
            Number of features in the input.
        hidden_size (int):
            Number of features in the hidden state.
        num_layers (int):
            Number of recurrent layers.
        output_size (int):
            Number of features in the output.
        batch_first (bool):
            If True, then the input and output tensors are provided as (batch,
            seq, feature). Default: True
        architecture (str):
            * 'RNN': RNN
            * 'GRU': GRU
            * 'LSTM': LSTM
        kwargs_architecture (dict):
            Extra keyword arguments for the architecture.
        nonlinearity (str):
            Nonlinearity to use. Can be either 'tanh' or 'relu'.
            Only used if architecture is RNN.
        bias (bool):
            If False, then the layer does not use bias weights b_ih and b_hh. Default: True
        dropout (float):
            If non-zero, introduces a Dropout layer on the outputs of each RNN layer except the last layer, with dropout probability equal to dropout. Default: 0
        bidirectional (bool):
            If True, becomes a bidirectional RNN. Default: False
        device (str):
            Device to use.
        dtype (torch.dtype):
            Data type to use.
    """
    def __init__(
        self, 
        features=1, 
        hidden_size=100, 
        num_layers=1, 
        output_size=1,
        batch_first=True,
        architecture='LSTM',
        kwargs_architecture={},
        nonlinearity='tanh',
        bias=True,
        dropout=0.0,
        bidirectional=False,
        device='cpu',
        dtype=torch.float32,
    ):
        super().__init__()
        self.input_size = features
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.output_size = output_size
        self.batch_first = batch_first
        self.architecture = architecture.upper()
        self.kwargs_architecture = kwargs_architecture
        self.nonlinearity = nonlinearity
        self.bias = bias
        self.dropout = dropout
        self.bidirectional = bidirectional
        self.device = device
        self.dtype = dtype

        assert hasattr(torch.nn, self.architecture), f'Architecture {self.architecture} not implemented'

        if self.architecture in ['RNN']:
            self.kwargs_architecture['nonlinearity'] = self.nonlinearity

        self.rnn = getattr(torch.nn, self.architecture)(
            input_size=self.input_size, 
            hidden_size=self.hidden_size, 
            num_layers=self.num_layers, 
            batch_first=self.batch_first,
            device=self.device,
            dtype=self.dtype,
            bias=self.bias,
            dropout=self.dropout,
            bidirectional=self.bidirectional,
            **self.kwargs_architecture,
        )
        self.fc = torch.nn.Linear(self.hidden_size * (2 if self.bidirectional else 1), self.output_size)

    def forward(self, x, hidden=None, hidden_initialization='zeros'):
        """
        Forward pass of the RNN

        Args:
            x (torch.Tensor):
                Input tensor of shape (batch_size, seq_len, input_size).
            hidden (torch.Tensor):
                Initial hidden state.
                If architecture is RNN or GRU: \
                    * Shape: (num_layers * (2 if bidirectional else 1), batch_size, hidden_size)
                If architecture is LSTM: \
                    * Shape: (num_layers * (2 if bidirectional else 1), batch_size, hidden_size)
                    * Tuple of (h0, c0)
            hidden_initialization (str):
                Initialization method used if hidden is None.
                    * 'zeros': Zeros
                    * 'ones': Ones
                    * 'random_normal': Random normal
                    * 'random_uniform': Random uniform
                    * 'xavier_normal': Xavier normal
                    * 'xavier_uniform': Xavier uniform
                    * 'kaiming_normal': Kaiming normal
                    * 'kaiming_uniform': Kaiming uniform
                    * 'orthogonal': Orthogonal
                                    
        Returns:
            out (torch.Tensor):
                Output tensor of shape (batch_size, seq_len, output_size).
        """
        if hidden is None:
            hidden = self.initialize_hidden(
                batch_size=x.size(0), 
                hidden_initialization=hidden_initialization,
            )
        out, hidden = self.rnn(x, hidden)  ## Note that hidden is a tuple for LSTM (h0, c0)
        out = self.fc(out)
        return out, hidden
    
    def initialize_hidden(self, batch_size, hidden_initialization='zeros'):
        """
        Initialize hidden state

        Args:
            batch_size (int):
                Batch size.
            hidden_initialization (str):
                Initialization method used if hidden is None.
                    * 'zeros': Zeros
                    * 'ones': Ones
                    * 'random_normal': Random normal
                    * 'random_uniform': Random uniform
                    * 'xavier_normal': Xavier normal
                    * 'xavier_uniform': Xavier uniform
                    * 'kaiming_normal': Kaiming normal
                    * 'kaiming_uniform': Kaiming uniform
                    * 'orthogonal': Orthogonal

        Returns:
            hidden (torch.Tensor):
                Initial hidden state.
                If architecture is RNN or GRU: \
                    * Shape: (num_layers * (2 if bidirectional else 1), batch_size, hidden_size)
                If architecture is LSTM: \
                    * Shape: (num_layers * (2 if bidirectional else 1), batch_size, hidden_size)
                    * Tuple of (h0, c0)
        """
        
        fn_empty = lambda : torch.empty(
            self.num_layers * (2 if self.rnn.bidirectional else 1),
            batch_size,
            self.hidden_size,
            device=self.device,
            dtype=self.dtype,
        )

        ## Make lambda functions that fill an empty tensor with the desired initialization
        if hidden_initialization == 'zeros':
            fn_fill = lambda e: torch.nn.init.zeros_(e)
        elif hidden_initialization == 'ones':
            fn_fill = lambda e: torch.nn.init.ones_(e)            
        elif hidden_initialization == 'random_normal':
            fn_fill = lambda e: torch.nn.init.normal_(e)
        elif hidden_initialization == 'random_uniform':
            fn_fill = lambda e: torch.nn.init.uniform_(e)
        elif hidden_initialization == 'xavier_normal':
            fn_fill = lambda e: torch.nn.init.xavier_normal_(e)
        elif hidden_initialization == 'xavier_uniform':
            fn_fill = lambda e: torch.nn.init.xavier_uniform_(e)
        elif hidden_initialization == 'kaiming_normal':
            fn_fill = lambda e: torch.nn.init.kaiming_normal_(e)
        elif hidden_initialization == 'kaiming_uniform':
            fn_fill = lambda e: torch.nn.init.kaiming_uniform_(e)
        elif hidden_initialization == 'orthogonal':
            fn_fill = lambda e: torch.nn.init.orthogonal_(e)
        else:
            raise ValueError(f'Unknown hidden_initialization {hidden_initialization}')
        
        ## Fill the empty tensor
        if self.architecture in ['RNN', 'GRU']:
            hidden = fn_fill(fn_empty())
        elif self.architecture == 'LSTM':
            hidden = (fn_fill(fn_empty()), fn_fill(fn_empty()))
        
        return hidden
    
    def to(self, device):
        self.device = device
        self.rnn.to(device)
        self.fc.to(device)
        return self
    

class Dataloader_MultiTimeSeries():
    """
    Dataloader for loading random chunks of multiple time series that all have
    the same number of features but may have different durations.
    The inputs are a list of 2D arrays of shape (duration, num_features).
    The output is a batch of shape (batch_size=n_datasets, batch_duration,
    num_features).
    Note that this is not a true pytorch Dataloader class, and does not support
    multiprocessing and other features.

    Args:
        timeseries (list):
            List of 2D arrays of shape (duration, num_features).
            Each array should be a torch.Tensor
            If 1D, then it will be converted to 2D with shape (duration, 1).
        batch_size (int):
            Batch size.
        batch_duration (int):
            Duration of each batch.
        shuffle_datasets (bool):
            If True, then the datasets are shuffled.
            If False, then the batch_size will be the number of datasets and the
            batch_size argument will be ignored.
        shuffle_startIdx (bool):
            If True, then the starting indices of each batch are shuffled each epoch.
            If False, then the sequence will be sequential chunks of
            batch_duration starting at 0.
        device (str):
            Device to use.
        """
    def __init__(
        self, 
        timeseries, 
        batch_size=1, 
        batch_duration=100, 
        shuffle_datasets=True, 
        shuffle_startIdx=True,
    ): 
        if not isinstance(timeseries, list):
            timeseries = [timeseries]
        
        assert all([isinstance(ts, torch.Tensor) for ts in timeseries]), 'All timeseries must be torch.Tensor'
        timeseries = [ts[:,None] if ts.ndim == 1 else ts for ts in timeseries]
        assert all([ts.ndim == 2 for ts in timeseries]), 'All timeseries must be 2D'
        assert all([ts.shape[1] == timeseries[0].shape[1] for ts in timeseries]), f"All timeseries must have the same number of features. Got {[ts.shape for ts in timeseries]}"
        
        self.batch_size = batch_size
        self.batch_duration = batch_duration
        self.shuffle_datasets = shuffle_datasets
        self.shuffle_startIdx = shuffle_startIdx

        self.num_datasets = len(timeseries)
        self.num_features = timeseries[0].shape[1]
        self.durations = [ts.shape[0] for ts in timeseries]
        self.duration_min = min(self.durations)

        self.timeseries = timeseries
        self.num_batches = self.duration_min // self.batch_duration
        self.idx_starts_raw = [
            torch.arange(0, d - self.batch_duration, self.batch_duration) 
            for d in self.durations
        ]

        self.idx_datasets_ = torch.arange(self.num_datasets)

        self.current_batch_ = 0
        self.idx_starts_ = None

    def __len__(self):
        return self.num_batches

    def __iter__(self):
        self.current_batch_ = 0
        
        if self.shuffle_startIdx:
            ## Roll (modulo % the duration - batch_duration) and randperm each idx_starts_raw
            rolls = [int(torch.randint(0, d, (1,))[0]) for d in self.durations]
            self.idx_starts_ = np.array([
                (torch.roll(self.idx_starts_raw[idx], rolls[idx]) % (self.durations[idx] - self.batch_duration))[torch.randperm(len(self.idx_starts_raw[idx]))]
                for idx in torch.arange(self.num_datasets)
            ], dtype=object)

            self.current_startIdx_ = torch.zeros(len(self.idx_starts_), dtype=torch.long)
        else:
            self.idx_starts_ = self.idx_starts_raw
            self.current_startIdx_ = torch.zeros(len(self.idx_starts_), dtype=torch.long)

        return self

    def _next_startIdx(self, idx_dataset):
        if self.current_startIdx_[idx_dataset].item() >= len(self.idx_starts_[idx_dataset]):
            raise StopIteration
        
        idx_start = self.idx_starts_[idx_dataset][self.current_startIdx_[idx_dataset]]
        self.current_startIdx_[idx_dataset] += 1
        return idx_start

    def __next__(self):
        if self.current_batch_ >= self.num_batches:
            raise StopIteration

        # Shuffle datasets and starting indices if required
        if self.shuffle_datasets:
            self.idx_datasets_ = torch.arange(self.num_datasets)[torch.randint(0, self.num_datasets, (self.batch_size,))]

        # Prepare batch data
        batch_data = [
            self.timeseries[idx_data][(w := self._next_startIdx(idx_data)):w + self.batch_duration,:]
            for idx_data in self.idx_datasets_
        ]

        self.current_batch_ += 1
        return torch.stack(batch_data, dim=0)
    
    def to(self, device):
        self.device = device
        self.timeseries = [ts.to(device) for ts in self.timeseries]
        return self


class BinaryClassification_RNN():
    def __init__(
        self,
        features=1,
        batch_size=50,
        batch_duration=900,
        num_epochs=1000,
        hidden_size=20,
        num_layers=2,
        dropout=0.0,
        architecture='RNN',
        val_check_period=10,
        X_val=None,
        y_val=None,
        device='cpu',
        lr=0.01,
        verbose=True,
    ):
        self.batch_size = batch_size
        self.batch_duration = batch_duration
        self.num_epochs = num_epochs

        self.X_val = X_val
        self.y_val = y_val

        self.device = device
        self.val_check_period=val_check_period
        self.verbose = verbose

        if (self.X_val is not None) and (self.y_val is not None):
            assert self.X_val.ndim in [2], f'X_val must be 2D. Got {self.X_val.ndim}D'
            assert self.y_val.ndim in [1], f'y_val must be 1D. Got {self.y_val.ndim}D'
            assert self.X_val.shape[0] == self.y_val.shape[0], f'X_val and y_val must have the same number of samples. Got {self.X_val.shape[0]} and {self.y_val.shape[0]}'
            self.X_val = torch.as_tensor(self.X_val, dtype=torch.float32, device=self.device)[None,:,:]
            self.y_val = torch.as_tensor(self.y_val, dtype=torch.float32, device=self.device)[None,:,None]
            self.run_validation = True
        else:
            self.run_validation = False
    
        self.model = RegressionRNN(
            features=features, 
            hidden_size=hidden_size, 
            num_layers=num_layers, 
            output_size=1,
            batch_first=True,
            architecture=architecture,
            kwargs_architecture={},
            nonlinearity='tanh',
            bias=True,
            dropout=dropout,
            bidirectional=False,
            device=self.device,
            dtype=torch.float32,
        ).to(self.device)
        
        # Loss and optimizer
        self.criterion = torch.nn.BCEWithLogitsLoss()
        self.optimizer = torch.optim.Adam(self.model.parameters(), lr=lr)
        self.loss_all = {}
        self.loss_val_all = {}
        self.epoch = 0

    def fit(
        self,
        X,
        y,
    ):
        if isinstance(X, list):
            assert all([x.ndim == 2 for x in X]), f'All X must be 2D. Got {[x.ndim for x in X]}'
            assert all([x.shape[1] == X[0].shape[1] for x in X]), f'All X must have the same number of features. Got {[x.shape for x in X]}'
            assert all([y.ndim == 1 for y in y]), f'All y must be 1D. Got {[y.ndim for y in y]}'
            timeseries = [torch.as_tensor(np.concatenate((y[:,None], x), axis=1), dtype=torch.float32, device=self.device) for x, y in zip(X, y)]
        else:
            assert isinstance(X, (np.ndarray, torch.Tensor)), f'X must be a list of 2D arrays or a 2D array. Got {type(X)}'
            assert X.ndim == 2, f'X must be 2D. Got {X.ndim}'
            assert X.shape[0] == y.shape[0], f'X and y must have the same number of samples. Got {X.shape[0]} and {y.shape[0]}'
            assert y.ndim == 1, f'y must be 1D. Got {y.ndim}'
            timeseries = [torch.as_tensor(np.concatenate((y[:,None], X), axis=1), dtype=torch.float32, device=self.device)]

        self.dataloader_XYcat = Dataloader_MultiTimeSeries(
            timeseries=timeseries,
            batch_size=self.batch_size,
            batch_duration=self.batch_duration,
            shuffle_datasets=True,
            shuffle_startIdx=True,
        )            
        
        num_epochs = self.epoch + self.num_epochs
        for self.epoch in tqdm(range(self.epoch, num_epochs), disable=not self.verbose):
            self.model.train()
            for iter_x, batch in enumerate(self.dataloader_XYcat):
                x_batch = batch[:,:,1:].to(self.device)
                y_batch = batch[:,:,0:1].to(self.device)
            
                self.optimizer.zero_grad()
                outputs, hidden = self.model.forward(
                    x=x_batch,
                    hidden_initialization='orthogonal',
                ) 
                # Ensure output dimensions match target dimensions
                loss = self.criterion(outputs.view(-1, 1), y_batch.view(-1,1))
                loss.backward()
                self.optimizer.step()
            
                self.loss_all[(self.epoch, iter_x)] = loss.item()
            
            if self.epoch >= num_epochs:
                break

            if self.run_validation and ((self.epoch % self.val_check_period) == 0):
                self.model.eval()
                y_hat_val, hidden = self.model.forward(
                    x=self.X_val,
                    hidden_initialization='orthogonal',
                )
                loss_val = self.criterion(y_hat_val.view(-1, 1), self.y_val.view(-1,1)).item()
                self.loss_val_all[self.epoch] = loss_val
            else:
                loss_val = torch.nan
                
            if self.verbose:
                print(f'Epoch [{self.epoch+1}/{num_epochs}], Loss: {loss.item():.4f}, Loss val: {loss_val:.4f}')

    def predict_proba(self, X):
        self.model.eval()
        y_hat, hidden = self.model.forward(
            x=torch.as_tensor(X, dtype=torch.float32, device=self.device)[None,:,:],
            hidden_initialization='orthogonal',
        )
        proba = torch.sigmoid(y_hat).detach().cpu().numpy()[0,:,0]
        proba_onehot = np.stack([1-proba, proba], axis=-1)
        return proba_onehot
    
    def predict(self, X):
        return np.argmax(self.predict_proba(X), axis=-1)
    
    def to(self, device):
        self.device = device
        self.model.to(device)
        
        ## Move optimizer parameters to new device
        for state in self.optimizer.state.values():
            for k, v in state.items():
                if isinstance(v, torch.Tensor):
                    state[k] = v.to(device)

        ## Move X_val and y_val to new device
        if hasattr(self, 'X_val'):
            if isinstance(self.X_val, torch.Tensor):
                self.X_val = self.X_val.to(device)
        if hasattr(self, 'y_val'):
            if isinstance(self.y_val, torch.Tensor):
                self.y_val = self.y_val.to(device)

        ## Move dataloader to new device
        if hasattr(self, 'dataloader_XYcat'):
            self.dataloader_XYcat.to(device)
            
        return self


import numpy as np
import sys
import re
import hashlib
from pathlib import Path
import warnings
from typing import Callable, List, Any, Dict
from contextlib import contextmanager


def estimate_array_size(
    array=None, 
    numel=None, 
    input_shape=None, 
    bitsize=64, 
    units='GB',
):
    '''
    Estimates the size of a hypothetical array based on shape or number of 
    elements and the bitsize
    RH 2021

    Args:
        numel (int): 
            number of elements in the array. If None, then 'input_shape'
            is used instead
        input_shape (tuple of ints):
            shape of array. Output of array.shape . Used if numel is None
        bitsize (int):
            bit size / width of the hypothetical data. eg:
                'float64'=64
                'float32'=32
                'uint8'=8
        units (str):
            units of the output. eg:
                'GB'=gigabytes
                'MB'=megabytes
                'KB'=kilobytes
                'B'=bytes
    
    Returns:
        size_estimate_in_bytes (int):
            size, in bytes, of hypothetical array. Doesn't include metadata,
            but for numpy arrays, this is usually very small (~128 bytes)

    '''
    ## Either array supplied or numel or input_shape
    assert sum([(array is not None), (numel is not None), (input_shape is not None)]) == 1, 'Exactly one of array, numel, or input_shape must be supplied'

    if array is not None:
        assert hasattr(array, 'shape'), f'array must have a shape attribute'
        input_shape = array.shape

        if isinstance(array, np.ndarray):
            bitsize = array.dtype.itemsize * 8
        else:
            import torch
            if isinstance(array, torch.Tensor):
                bitsize = array.element_size() * 8
            else:
                raise TypeError(f'array must be a numpy or torch array. Got {type(array)}')

    if numel is None:
        numel = np.prod(np.array(input_shape, dtype=np.float64))
    
    bytes_per_element = float(bitsize / 8)
    
    size_estimate_in_bytes = numel * bytes_per_element
    size_out = convert_size(size_estimate_in_bytes, units)
    return size_out


def get_vars(globals, size_thresh=0, var_type=None, return_vars_pref=False):
    '''
    Returns variable info that matches defined criteria.
    RH 2021

    Args:
        globals:
            `globals()` must be passed here
        size_thresh (scalar):
            Minimum size, in MB of variables you'd like returned
        var_type (type obj):
            Class type you'd like returned
        return_vars_pref (bool):
            Whether or not you'd like the outputs returned

    Returns:
        var_names (np.array of str):
            Names of variables sorted by size
        var_sizes (np.array of float64):
            Sizes of variables sorted by size
        var_types (np.array of type objects):
            Types of variables sorted by size

    Demo:
        var_names, var_sizes, var_types = get_vars(globals(), size_thresh=0.1, var_type=np.ndarray, return_vars_pref=True)
    '''
    var_info = []
    for ii, (name, var) in enumerate(globals.items()):
        var_info.append((name, sys.getsizeof(var), type(var)))

    var_names = np.array(var_info)[:,0]
    var_sizes = np.float64(np.array(var_info)[:,1])
    var_types = np.array(var_info)[:,2]

    sort_idx = np.flip(np.argsort(var_sizes))
    var_types = var_types[sort_idx]
    var_names = var_names[sort_idx]
    var_sizes = var_sizes[sort_idx]

    idx_toInclude = []
    for ii, (name, size, val_type) in enumerate(zip(var_names, var_sizes, var_types)):
        if var_type is not None:
            if size > size_thresh*1000000 and (val_type==var_type):
                idx_toInclude.append(ii)
                print(f'{name}, {size/1000000} MB, type: {val_type}')
        else:
            if size > size_thresh*1000000:
                idx_toInclude.append(ii)
                print(f'{name}, {size/1000000} MB, type: {val_type}')
    
    if return_vars_pref:
        return var_names[idx_toInclude], var_sizes[idx_toInclude], var_types[idx_toInclude]


def get_nums_from_str(str_in, dtype_out=np.float64):
    """
    Returns a list of numbers from a string.
    Numbers can be negative and decimals.
    RH 2022

    Args:
        str_in (str):
            String to be parsed.
            Should contain numbers separated by spaces, commas,
             letters, or most other characters.
        dtype_out (type obj):
            dtype of output.

    Returns:
        nums (np.array):
            List of numbers found in the string.

    """
    return np.array([float(i) for i in re.findall(r'\-?\d+\.?\d*', str_in)], dtype=dtype_out)


def write_to_log(
    text, 
    path_log, 
    mode='a', 
    start_on_new_line=True, 
    pref_print=True, 
    pref_save=True
):
    """
    Writes text to a log file.
    RH 2022

    Args:
        path_log (str):
            Path to log file.
        text (str):
            Text to write to log file.
        mode (str):
            Mode to open log file in.
            'a' for append, 'w' for write.
        start_on_new_line (bool):
            Whether or not to start on a new line.
        pref_print (bool):
            Whether or not to print text to console.
        pref_save (bool):
            Whether or not to save text to log file.

    Returns:
        None
    """
    if pref_print:
        print(text)
    if pref_save:
        with open(path_log, mode=mode) as log:
            if start_on_new_line==True:
                log.write('\n')
            log.write(str(text))


def format_text(
    text, 
    color=None, 
    bold=False, 
    italic=False, 
    background=None, 
    underline=False, 
    blink=False, 
    dim=False,
):
    """
    Formats text with ANSI escape sequences.
    RH 2023

    Args:
        text (str):
            Text to format.
        color (tuple of ints):
            RGB color of text.
            Range: 0-255
        bold (bool):
            Whether or not to bold text.
        italic (bool):
            Whether or not to italicize text.
            Doesn't work in ipynb
        background (tuple of ints):
            RGB color of text background.
            Range: 0-255
        underline (bool):
            Whether or not to underline text.
        blink (bool):
            Whether or not to blink text.
            Doesn't work in ipynb
        dim (bool):
            Whether or not to dim text.
            Doesn't work in ipynb

    Returns:
        formatted_text (str):
            Formatted text.
    """
    # ANSI escape sequences for text formatting
    modifiers = []
    
    # Color
    if color:
        r, g, b = color
        modifiers.append(f'\x1b[38;2;{r};{g};{b}m')

    # Background color
    if background:
        r, g, b = background
        modifiers.append(f'\x1b[48;2;{r};{g};{b}m')
    
    # Other modifiers
    if bold:
        modifiers.append('\x1b[1m')
    if italic:
        modifiers.append('\x1b[3m')
    if underline:
        modifiers.append('\x1b[4m')
    if blink:
        modifiers.append('\x1b[5m')
    if dim:
        modifiers.append('\x1b[2m')

    # Reset all formatting
    reset = '\x1b[0m'
    
    return ''.join(modifiers) + text + reset


def system_info(verbose: bool = False,) -> Dict:
    """
    Checks and prints the versions of various important software packages.
    RH 2022

    Args:
        verbose (bool): 
            Whether to print the software versions. 
            (Default is ``False``)

    Returns:
        (Dict): 
            versions (Dict):
                Dictionary containing the versions of various software packages.
    """
    ## Operating system and version
    import platform
    def try_fns(fn):
        try:
            return fn()
        except:
            return None
    fns = {key: val for key, val in platform.__dict__.items() if (callable(val) and key[0] != '_')}
    operating_system = {key: try_fns(val) for key, val in fns.items() if (callable(val) and key[0] != '_')}
    print(f'== Operating System ==: {operating_system["uname"]}') if verbose else None

    ## CPU info
    try:
        import cpuinfo
        import multiprocessing as mp
        # cpu_info = cpuinfo.get_cpu_info()
        cpu_n_cores = mp.cpu_count()
        cpu_brand = cpuinfo.cpuinfo.CPUID().get_processor_brand(cpuinfo.cpuinfo.CPUID().get_max_extension_support())
        cpu_info = {'n_cores': cpu_n_cores, 'brand': cpu_brand}
        if 'flags' in cpu_info:
            cpu_info['flags'] = 'omitted'
    except Exception as e:
        warnings.warn(f'RH WARNING: unable to get cpu info. Got error: {e}')
        cpu_info = 'Error: Failed to get'
    print(f'== CPU Info ==: {cpu_info}') if verbose else None

    ## RAM
    import psutil
    ram = psutil.virtual_memory()
    print(f'== RAM ==: {ram}') if verbose else None

    ## User
    import getpass
    user = getpass.getuser()

    ## GPU
    try:
        import GPUtil
        gpus = GPUtil.getGPUs()
        gpu_info = {gpu.id: gpu.__dict__ for gpu in gpus}
    except Exception as e:
        warnings.warn(f'RH WARNING: unable to get gpu info. Got error: {e}')
        gpu_info = 'Error: Failed to get'
    print(f'== GPU Info ==: {gpu_info}') if verbose else None
    
    ## Conda Environment
    import os
    if 'CONDA_DEFAULT_ENV' not in os.environ:
        conda_env = 'None'
    else:
        conda_env = os.environ['CONDA_DEFAULT_ENV']
    print(f'== Conda Environment ==: {conda_env}') if verbose else None

    ## Python
    import sys
    python_version = sys.version.split(' ')[0]
    print(f'== Python Version ==: {python_version}') if verbose else None

    ## GCC
    import subprocess
    try:
        gcc_version = subprocess.check_output(['gcc', '--version']).decode('utf-8').split('\n')[0].split(' ')[-1]
    except Exception as e:
        warnings.warn(f'RH WARNING: unable to get gcc version. Got error: {e}')
        gcc_version = 'Faled to get'
    print(f'== GCC Version ==: {gcc_version}') if verbose else None
    
    ## PyTorch
    import torch
    torch_version = str(torch.__version__)
    print(f'== PyTorch Version ==: {torch_version}') if verbose else None
    ## CUDA
    if torch.cuda.is_available():
        cuda_version = torch.version.cuda
        cudnn_version = torch.backends.cudnn.version()
        torch_devices = [f'device {i}: Name={torch.cuda.get_device_name(i)}, Memory={torch.cuda.get_device_properties(i).total_memory / 1e9} GB' for i in range(torch.cuda.device_count())]
        print(f"== CUDA Version ==: {cuda_version}, CUDNN Version: {cudnn_version}, Number of Devices: {torch.cuda.device_count()}, Devices: {torch_devices}, ") if verbose else None
    else:
        cuda_version = None
        cudnn_version = None
        torch_devices = None
        print('== CUDA is not available ==') if verbose else None

    ## all packages in environment
    import pkg_resources
    pkgs_dict = {i.key: i.version for i in pkg_resources.working_set}

    ## get datetime
    from datetime import datetime
    dt = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

    versions = {
        'datetime': dt,
        'operating_system': operating_system,
        'cpu_info': cpu_info,  ## This is the slow one.
        'user': user,
        'ram': ram,
        'gpu_info': gpu_info,
        'conda_env': conda_env,
        'python': python_version,
        'gcc': gcc_version,
        'torch': torch_version,
        'cuda': cuda_version,
        'cudnn': cudnn_version,
        'torch_devices': torch_devices,
        'pkgs': pkgs_dict,
    }

    return versions


def array_hasher():
    """
    Returns a function that hashes an array.
    """
    from functools import partial
    import xxhash
    return partial(xxhash.xxh64_hexdigest, seed=0)


@contextmanager
def temp_set_attr(obj, attr_name, new_value):
    """
    Temporarily set an attribute of an object to a new value within a context
    manager / closure.
    RH 2024

    Args:
        obj (object):
            Object to toggle attribute for.
        attr_name (str):
            Attribute to toggle.
        new_value (Any):
            New value to set attribute to.

    Demo:
        .. code-block:: python
            
                with temp_set_attr(obj, attr, new_val):
                    # do something
    """
    original_value = getattr(obj, attr_name)
    setattr(obj, attr_name, new_value)
    try:
        yield
    finally:
        setattr(obj, attr_name, original_value)


def wrapper_flexible_args(names_kwargs: List[str]) -> Callable:
    """
    Make a wrapper function that allows for flexible argument names.\n
    Useful for things like `dim` vs. `axis` or `keepdim` vs. `keepdims`.\n
    RH 2024

    Args:
        names_kwargs (List[str]):
            List of argument names that the wrapper function should accept.

    Returns:
        (Callable):
            Wrapper function.
    """
    import inspect
    from functools import wraps

    def wrapper(func: Callable) -> Callable:
        try:
            sig = inspect.signature(func)
        except ValueError:
            raise ValueError(f'Function {func.__name__} must have a signature to use this wrapper')
        
        @wraps(func)
        def wrapped(*args, **kwargs):
            ## Check for intersection of names_kwargs with sig.parameters and kwargs
            sp = set(sig.parameters.keys())
            nk = set(names_kwargs)
            k = set(kwargs.keys())
            ix_nk_sp = nk.intersection(sp)
            ix_nk_k = nk.intersection(k)
            ix_nk_k_sp = ix_nk_k.intersection(sp)
            ## Logic steps:
            ### If none of the kwargs are in the signature, and there is at least one named_kwarg in both the kwargs and signature, then rename the kwargs
            if (len(ix_nk_k_sp) == 0) and (len(ix_nk_k) > 0) and (len(ix_nk_sp) > 0):
                k_toUse = ix_nk_sp.pop()
                k_toRename = ix_nk_k.pop()
                kwargs[k_toUse] = kwargs.pop(k_toRename)
            return func(*args, **kwargs)
        return wrapped
    return wrapper


#########################################################
############ INTRA-MODULE HELPER FUNCTIONS ##############
#########################################################


def convert_size(size, return_size='GB'):
    """
    Convert size to GB, MB, KB, from B.
    RH 2021

    Args:
        size (int or float):
            Size in bytes.
        return_size (str):
            Size unit to return.
            Options: 'TB', 'GB', 'MB', or 'KB'
        
    Returns:
        out_size (float):
            Size in specified unit.      
    """

    if return_size == 'TB':
        out_size = size / 1000000000000
    elif return_size == 'GB':
        out_size = size / 1000000000
    elif return_size == 'MB':
        out_size = size / 1000000
    elif return_size == 'KB':
        out_size = size / 1000
    elif return_size == 'B':
        out_size = size / 1

    return out_size



# def align_dtw(alignment_index1, alignment_index2 , queryTrace):
# #     ```
# #     alignmentObj: alignment object from dtw-python call
# #     queryTrace: the query (warped) variable from the dtw-python call
# #     ```
#     import numpy as np
#     idx_multi_bool = np.diff(alignment_index2)==0
#     idx_multi_bool = np.hstack([idx_multi_bool , 0]) + np.hstack([0 , idx_multi_bool])
#     idx_multi = alignment_index2[np.where(idx_multi_bool)[0]]

#     output = np.zeros(len(alignment_index2)) * np.nan
#     for ii in range(len(alignment_index2)):
#         if idx_multi_bool[ii]:
#             tmp_idx_multi = alignment_index2==alignment_index2[ii]
#             output[alignment_index2[ii]] = np.mean(queryTrace[alignment_index1[tmp_idx_multi]])
#         else:
#             output[alignment_index2[ii]] = queryTrace[alignment_index1[ii]]
#     output = output[np.isnan(output)==0]
#     return output

# import dtw
# from tqdm import trange
# import numpy as np

# ## traces need to be pretty similar to use this code

# ## this dtw library is poorly documented and kind of janky. It won't let me put in large sequences. Anything over
# # 40k samples takes up 100s of GB of memory. f that. so I wrote a thing to split it up into chunks. There are some
# # edge effects, so the chunking also has some overlaps that are removed.

# ## The main outputs here are the matched indices: alignment_index1 and alignment_index2. Use these with the align_dtw
# # function to warp a signal

# input_template = scipy.signal.savgol_filter(np.diff(X_cc[:,0]) , 31,3)
# input_query = -scipy.signal.savgol_filter(np.diff(X_cc[:,1]) , 31,3)

# last_good_idx_query = int(np.max(np.where(np.isnan(input_query)==0)[0]))

# template = input_template[:last_good_idx_query]
# query = input_query[:last_good_idx_query]

# chunk_size = 10000
# overlap_size = 1000 # make even and the length of the window of possible shifts

# num_chunks = int(np.ceil(len(template)/chunk_size))

# alignment_list_extended = list(np.ones(num_chunks))
# alignment_list_index1 = list(np.ones(num_chunks))
# alignment_list_index2 = list(np.ones(num_chunks))
# for ii in trange(num_chunks):
#     idx_chunk_extended = np.arange(np.maximum(ii*chunk_size - overlap_size , 0), 
#                                    np.minimum( (ii+1)*chunk_size + overlap_size , len(template)))
#     template_chunk = template[idx_chunk_extended]
#     query_chunk = query[idx_chunk_extended]
#     alignment_list_extended[ii] = dtw.dtw(query_chunk , template_chunk, 
#                     keep_internals=True,
#                    step_pattern=dtw.rabinerJuangStepPattern(6,"c"),
# #                     window_type="sakoechiba", window_args={'window_size':10000}
#                    )
    
#     if ii==0:
#         idx_chunk = np.arange(0, np.min(np.where(alignment_list_extended[ii].index2==chunk_size)[0]))
#     elif ii==num_chunks-1:
#         idx_chunk = np.arange( np.min(np.where(alignment_list_extended[ii].index2==overlap_size)[0]),
#                       len(alignment_list_extended[ii].index2))
#     else:
#         idx_chunk = np.arange( np.min(np.where(alignment_list_extended[ii].index2==overlap_size)[0]),
#                               np.min(np.where(alignment_list_extended[ii].index2==chunk_size+overlap_size)[0]))
#     alignment_list_index1[ii] = alignment_list_extended[ii].index1[idx_chunk]
#     alignment_list_index2[ii] = alignment_list_extended[ii].index2[idx_chunk]

# alignment_index1 = []
# alignment_index2 = []
# for ii in range(num_chunks):
#     if ii==0:
#         alignment_index1 = np.uint64(np.hstack(( alignment_index1 ,
#                                             (alignment_list_index1[ii] + ii*chunk_size)) ))
#         alignment_index2 = np.uint64(np.hstack(( alignment_index2 ,
#                                             (alignment_list_index2[ii] + ii*chunk_size)) ))
#     else:
#         alignment_index1 = np.uint64(np.hstack(( alignment_index1 ,
#                                                 (alignment_list_index1[ii] + ii*chunk_size - overlap_size)) ))
#         alignment_index2 = np.uint64(np.hstack(( alignment_index2 ,
#                                                 (alignment_list_index2[ii] + ii*chunk_size - overlap_size)) ))


# ######
# np.corrcoef(template , output)[0,1]**2
# ######
# output = align_dtw(alignment_index1, alignment_index2 , query)

# plt.figure()
# plt.plot(alignment_index1)
# plt.plot(alignment_index2)

# plt.figure()
# # plt.plot(scipy.signal.savgol_filter(np.diff(alignment_index1) , 301,3))
# # plt.plot(scipy.signal.savgol_filter(np.diff(alignment_index2) , 301,3))
# # plt.plot(np.array(alignment_index1 , dtype='float64') - np.array(alignment_index2 , dtype='float64'))
# plt.plot(alignment_index1 , np.array(alignment_index1 , dtype='float64') - np.array(alignment_index2 , dtype='float64'))

# plt.figure()
# plt.plot(template)
# plt.plot(query)
# plt.plot(output)
# # plt.plot(scipy.signal.savgol_filter(np.diff(alignment_index1) , 301,3) - scipy.signal.savgol_filter(np.diff(alignment_index2) , 301,3))
# plt.plot(alignment_index1 , (np.array(alignment_index1 , dtype='float64') - np.array(alignment_index2 , dtype='float64')) / 30)
# plt.plot(X_cc[:,0]/10)


# # decoder_output_S2pWarped = align_dtw(alignment_index1, alignment_index2 , decoder_output)



'''
Table of Contents

Functions and Interdependencies:
    proj
    orthogonalize
        - proj
    OLS
    EV
    pairwise_similarity
    best_permutation
        - pairwise_similarity
    self_similarity_pairwise
        - best_permutation
'''
import numpy as np
import scipy.optimize
# import sklearn.decomposition
from numba import njit, prange, jit
import torch
from tqdm import tqdm

from . import indexing

import copy
import time
from functools import partial

def proj(v1, v2):
    '''
    Projects one or more vectors (columns of v1) onto one or more vectors (v2)
    RH 2021

    Args:
        v1 (ndarray): 
            vector set 1. Either a single vector or a 2-D array where the columns
             are the vectors
        v2 (ndarray): 
            vector set 2. Either a single vector or a 2-D array where the columns
             are the vectors
            If None, v2 is set to v1
    
    Returns:
        proj_vec (ndarray): 
            vector set 1 projected onto vector set 2. 
            shape: (v1.shape[0], v1.shape[1], v2.shape[1])
        proj_score (ndarray or scalar): 
            projection scores. shape: (v1.shape[1], v2.shape[1])
    '''
    if isinstance(v1, np.ndarray):
        from opt_einsum import contract
        einsum = contract
        norm = partial(np.linalg.norm, axis=0, keepdims=True)
    elif isinstance(v1, torch.Tensor):
        from torch import einsum
        norm = partial(torch.linalg.norm, dim=0, keepdim=True)

    if v2 is None:
        v2 = v1
    if v1.ndim < 2:
        v1 = v1[:,None]
    if v2.ndim < 2:
        v2 = v2[:,None]

    assert v1.shape[0] == v2.shape[0], f"v1.shape[0] must equal v2.shape[0]. Got v1.shape[0]={v1.shape[0]} and v2.shape[0]={v2.shape[0]}"

    u = v2 / norm(v2)
    proj_score = v1.T @ u 
    proj_vec = einsum('ik,jk->ijk', u, proj_score)

    return proj_vec, proj_score

def vector_angle(v1, v2=None, mode='cosine_similarity'):
    '''
    Calculates the angle between two vector sets.
    RH 2023

    Args:
        v1 (ndarray):
            vector set 1. Either a single vector or a 2-D array where the columns
             are the vectors
        v2 (ndarray):
            vector set 2. Either a single vector or a 2-D array where the columns
             are the vectors
            If None, v2 is set to v1
        mode (str):
            'deg': returns angle in degrees
            'rad': returns angle in radians
            'cosine_similarity': returns cosine similarity

    Returns:
        angle (ndarray or scalar):
            angle between v1 and v2. shape: (v1.shape[1], v2.shape[1])
    '''
    if isinstance(v1, np.ndarray):
        norm = partial(np.linalg.norm, axis=0, keepdims=True)
        arccos = np.arccos
        rad2deg = np.rad2deg
    elif isinstance(v1, torch.Tensor):
        norm = partial(torch.linalg.norm, dim=0, keepdim=True)
        arccos = torch.acos
        rad2deg = torch.rad2deg

    if v2 is None:
        v2 = v1
    if v1.ndim < 2:
        v1 = v1[:,None]
    if v2.ndim < 2:
        v2 = v2[:,None]

    assert v1.shape[0] == v2.shape[0], f"v1.shape[0] must equal v2.shape[0]. Got v1.shape[0]={v1.shape[0]} and v2.shape[0]={v2.shape[0]}"
    
    u1 = v1 / norm(v1)
    u2 = v2 / norm(v2)
    
    if mode == 'cosine_similarity':
        return u1.T @ u2
    elif mode == 'deg':
        return rad2deg(arccos(u1.T @ u2))
    elif mode == 'rad':
        return arccos(u1.T @ u2)
    else:
        raise ValueError(f"mode must be 'cosine_similarity', 'deg', or 'rad'. Got {mode}")

def orthogonalize(v1, v2, method='OLS', device='cpu', thresh_EVR_PCA=1e-15):
    '''
    Orthogonalizes one or more vectors (columns of v1) relative to another set 
     of vectors (v2). Subtracts the projection of v1 onto v2 off of v1.
    RH 2021-2023

    Args:
        v1 (ndarray): 
            vector set 1. (y_true)
            Either a single vector or a 2-D array where the columns
             are the vectors. shape: (n_samples, n_vectors)
        v2 (ndarray): 
            vector set 2. (y_pred)
            Either a single vector or a 2-D array where the columns
             are the vectors. shape: (n_samples, n_vectors)
        method (str):
            'serial': uses a gram-schmidt like iterative process to orthogonalize
             v1 off of v2. This method may have some minor numerical instability 
             issues when v2 contains many (>100) vectors.
            'OLS': uses OLS regression to orthogonalize v1 off of v2. This method 
             is numerically stable when v2 has many columns, and usually faster.
             However, OLS can have issues when v2 is not full rank or singular.
        device (str):
            Device to use for torch tensors. 
            Default is 'cpu'.
        thresh_EVR_PCA (scalar):
            Threshold for the Explained Variance Ratio (EVR) of the PCA components.
            Set so that it might act as a cutoff for redundant components.
            If the EVR of a component is below this threshold, it is not used to
             orthogonalize v1. This is useful when v2 is singular or nearly singular.
    
    Returns:
        v1_orth (ndarray): 
            vector set 1 with the projections onto vector set 2 subtracted off.
             Same size as v1.
        EVR (ndarray): 
            Explained Variance Ratio for each column of v1. 
            Amount of variance that all the vectors in v2 can explain for each 
            vector in v1.
            When v1 is z-scored, EVR is equivalent to pearsons R^2; 
            as in pairwise_similarity(OLS(v2, v1)[1] , v1)**2
        EVR_total (scalar):
            total amount of variance explained in v1 by v2
        pca_dict (dict):
            dictionary containing the PCA outputs:
                'comps': PCA components
                'scores': PCA scores
                'singVals': PCA singular values
                'EVR': PCA Explained Variance Ratio
                'PCs_above_thresh': boolean array indicating which PCs are above
                 the threshold for EVR
    '''
    if isinstance(v1, np.ndarray):
        v1 = torch.as_tensor(v1)
        return_numpy = True
    else:
        return_numpy = False
    if isinstance(v2, np.ndarray):
        v2 = torch.as_tensor(v2)

    v1 = v1.to(device)
    v2 = v2.to(device)

    if v1.ndim < 2:
        v1 = v1[:,None]
    if v2.ndim < 2:
        v2 = v2[:,None]
    
    # I'm pretty sure using PCA is fine for this.
    if v2.shape[1] > 1:
        from .decomposition import torch_pca
        comps, pc_scores, singVals, pc_EVR = torch_pca(
            X_in=v2, 
            rank=v2.shape[1], 
            mean_sub=True,  
            device=device, 
            return_cpu=False, 
            return_numpy=isinstance(v1, np.ndarray), 
            cuda_empty_cache=False
        )
        pca_dict = {
            'comps': comps,
            'scores': pc_scores,
            'singVals': singVals,
            'EVR': pc_EVR,
            'PCs_above_thresh': pc_EVR > thresh_EVR_PCA,
        }
        pc_scores_aboveThreshold = pcsat = pc_scores[:, pca_dict['PCs_above_thresh']]
    else:
        pcsat = v2
        pca_dict = None

    # decomp = sklearn.decomposition.PCA(n_components=v2.shape[1])
    # pc_scores = decomp.fit_transform(v2)

    v1_means = torch.mean(v1, dim=0)
    if method == 'serial':
        # Serial orthogonalization.
        v1_orth = copy.deepcopy(v1 - v1_means)
        for ii in range(pcsat.shape[1]):
            proj_vec = proj(v1_orth , pcsat[:,ii])[0]
            v1_orth = v1_orth.squeeze() - proj_vec.squeeze()
    elif method == 'OLS':
        # Ordinary Least Squares.
        X = torch.cat((pcsat, torch.ones((pcsat.shape[0], 1), dtype=pcsat.dtype, device=device)), dim=1)
        theta = torch.linalg.inv(X.T @ X) @ X.T @ (v1 - v1_means)
        y_rec = X @ theta
        v1_orth = v1 - y_rec + X[:,-1][:,None] * theta[-1]  ## this adds back the bias term

    v1_orth = v1_orth + v1_means

    EVR = 1 - (torch.var(v1_orth, dim=0) / torch.var(v1, dim=0))
    EVR_total = 1 - ( torch.sum(torch.var(v1_orth, dim=0), dim=0) / torch.sum(torch.var(v1, dim=0), dim=0) )

    if return_numpy:
        v1_orth = v1_orth.cpu().numpy()
        EVR = EVR.cpu().numpy()
        EVR_total = EVR_total.cpu().numpy()

        if pca_dict is not None:
            for key in pca_dict.keys():
                if isinstance(pca_dict[key], torch.Tensor):
                    pca_dict[key] = pca_dict[key].cpu().numpy()

    return v1_orth, EVR, EVR_total, pca_dict


@njit
def pair_orth_helper(v1, v2):
    """
    Helper function for main pairwise_orthogonalization
     function. Performs the pairwise orthogonalization
     by subtracting off v2 from v1. Uses numba to speed
     up the computation.
     v1 = v1 - proj(v1 onto v2)
    RH 2021

    Args:
        v1 (ndarray):
            Vector set 1. Columns are vectors.
        v2 (ndarray):
            Vector set 2. Columns are vectors.
    
    Returns:
        v1_orth (ndarray):
            v1 - proj(v1 onto v2)
    """
    return  v1 - (np.diag(np.dot(v1.T, v2)) / np.diag(np.dot(v2.T, v2))) * v2
def pairwise_orthogonalization(v1, v2, center:bool=True):
    """
    Orthogonalizes columns of v2 off of the columns of v1
     and returns the orthogonalized v1 and the explained
     variance ratio of v2 off of v1.
    v1: y_true, v2: y_pred
    Since it's just pairwise, there should not be any 
     numerical instability issues.
    RH 2021

    Args:
        v1 (ndarray):
            y_true
            Vector set 1. Either a single vector or a 2-D 
             array where the columns are the vectors.
        v2 (ndarray):  
            y_pred
            Vector set 2. Either a single vector or a 2-D
             array where the columns are the vectors.
        center (bool):
            Whether to center the vectors.
            Centering prevents negative EVR values.
    
    Returns:
        v1_orth (ndarray):
            Vector set 1 with the projections onto vector 
             set 2 subtracted off.
            Same size as v1.
        EVR (ndarray):
            Explained Variance Ratio for each column of v1.
            Amount of variance that all the vectors in v2 
             can explain for each vector in v1.
        EVR_total_weighted (scalar):
            Average amount of variance explained in v1 by v2
             weighted by the variance of each column of v1.
        EVR_total_unweighted (scalar):
            Average amount of variance explained in v1 by v2
    """
    assert v1.ndim == v2.ndim
    if v1.ndim==1:
        v1 = v1[:,None]
        v2 = v2[:,None]
    assert v1.shape[1] == v2.shape[1]
    assert v1.shape[0] == v2.shape[0]
    
    if center:
        v1 = v1 - np.mean(v1, axis=0)
        v2 = v2 - np.mean(v2, axis=0)
    
    v1_orth = pair_orth_helper(v1, v2)

    v1_var = np.var(v1, axis=0)
    EVR = 1 - (np.var(v1_orth, axis=0) / v1_var)

    EVR_total_weighted = np.sum(v1_var * EVR) / np.sum(v1_var)
    EVR_total_unweighted = np.mean(EVR)
    return v1_orth, EVR, EVR_total_weighted, EVR_total_unweighted


@torch.jit.script
def pairwise_orthogonalization_torch_helper(v1, v2, center:bool=True):
    # assert v1.ndim == v2.ndim
    if v1.ndim==1:
        v1 = v1[:,None]
    if v2.ndim==1:
        v2 = v2[:,None]
    # assert v1.shape[1] == v2.shape[1]
    assert v1.shape[0] == v2.shape[0]
    
    if center:
        v1 = v1 - torch.nanmean(v1, dim=0)
        v2 = v2 - torch.nanmean(v2, dim=0)

    # v1_orth = v1 - (torch.diag(torch.matmul(v1.T, v2)) / torch.diag(torch.matmul(v2.T, v2)))*v2
    v1_orth = v1 - (torch.nansum(v1 * v2, dim=0) / torch.nansum(v2 * v2, dim=0) )*v2

    v1_var = torch.var(v1, dim=0)
    EVR = 1 - (torch.var(v1_orth, dim=0) / v1_var)

    EVR_total_weighted = torch.nansum(v1_var * EVR) / torch.sum(v1_var)
    EVR_total_unweighted = torch.nanmean(EVR)
    return v1_orth, EVR, EVR_total_weighted, EVR_total_unweighted
def pairwise_orthogonalization_torch(v1, v2, center:bool=True, device='cpu'):
    """
    Orthogonalizes columns of v2 off of the columns of v1
     and returns the orthogonalized v1 and the explained
     variance ratio of v2 off of v1.
    Use center=True to get guarantee zero correlation between
     columns of v1_orth and v2.
    v1: y_true, v2: y_pred
    Since it's just pairwise, there should not be any 
     numerical instability issues.
    Same as pairwise_orthogonalization, but with 
     torch.jit.script instead of njit.
    RH 2021

    Args:
        v1 (ndarray):
            y_true
            Vector set 1. Either a single vector or a 2-D 
             array where the columns are the vectors.
        v2 (ndarray):  
            y_pred
            Vector set 2. Either a single vector or a 2-D
             array where the columns are the vectors.
        center (bool):
            Whether to center the vectors.
            Centering prevents negative EVR values.
    
    Returns:
        v1_orth (ndarray):
            Vector set 1 with the projections onto vector 
             set 2 subtracted off.
            Same size as v1.
        EVR (ndarray):
            Explained Variance Ratio for each column of v1.
            Amount of variance that all the vectors in v2 
             can explain for each vector in v1.
        EVR_total_weighted (scalar):
            Average amount of variance explained in v1 by v2
             weighted by the variance of each column of v1.
        EVR_total_unweighted (scalar):
            Average amount of variance explained in v1 by v2
    """
    if isinstance(v1, np.ndarray):
        v1 = torch.from_numpy(v1)
        return_numpy = True
    else:
        return_numpy = False
    if isinstance(v2, np.ndarray):
        v2 = torch.from_numpy(v2)
    v1 = v1.to(device)
    v2 = v2.to(device)
    v1_orth, EVR, EVR_total_weighted, EVR_total_unweighted = pairwise_orthogonalization_torch_helper(v1, v2, center=center)
    if return_numpy:
        v1_orth = v1_orth.cpu().numpy()
        EVR = EVR.cpu().numpy()
        EVR_total_weighted = EVR_total_weighted.cpu().numpy()
        EVR_total_unweighted = EVR_total_unweighted.cpu().numpy()
    return v1_orth, EVR, EVR_total_weighted, EVR_total_unweighted

def EV(y_true, y_pred):
    '''
    Explained Variance
    Calculating as 1 - (SS_residual_variance / SS_original_variance)
    Should be exactly equivalent to sklearn's 
    sklearn.metrics.explained_variance_score but slightly faster and provides
    most relevant outputs all together
    RH 2021

    Args:
        y_true (ndarray): 
            array with same size as y_pred. The columns are
            individual y vectors
        y_pred (ndarray):
            array with same size as y_true. The columns are the predicted
            y vectors
    
    Returns:
        EV (1-D array):
            explained variance of each y_true column by each corresponding
            y_pred column. Same as 
            sklearn.metrics.explained_variance_score(y_true, y_pred, multioutput='raw_values')
        EV_total_weighted (scalar):
            total weighted explained variance. Same as
            sklearn.metrics.explained_variance_score(y_true, y_pred, multioutput='variance_weighted')
        EV_total_unweighted (scalar):
            average of all EV values. Same as
            sklearn.metrics.explained_variance_score(y_true, y_pred, multioutput='uniform_average')
    '''

    EV = 1 - np.sum((y_true - y_pred)**2, axis=0) / np.sum((y_true - np.mean(y_true, axis=0))**2, axis=0)
    y_true_var = np.var(y_true, axis=0)
    EV_total_weighted = np.sum( y_true_var* EV ) / np.sum(y_true_var)
    EV_total_unweighted = np.mean(EV)

    return EV , EV_total_weighted , EV_total_unweighted


def pairwise_similarity(v1 , v2=None , method='pearson' , ddof=1):
    '''
    Computes similarity matrices between two sets of vectors (columns within
    2-D arrays) using either covariance, Pearson correlation or cosine_similarity.
    
    Think of this function as a more general version of np.cov or np.corrcoef
    RH 2021

    Args:
        v1 (ndarray): 
            2-D array of column vectors.
        v2 (ndarray): 
            2-D array of column vectors to compare to vector_set1. If None, then
             the function is a type of autosimilarity matrix
        method (str):
            'cov' - covariance
            'pearson' or 'R' - Pearson correlation
            'cosine_similarity' - cosine similarity
        ddof (scalar/int):
            Used if method=='cov'. Define the degrees of freedom. Set to 1 for
            unbiased calculation, 0 for biased calculation.
    Returns:
        ouput (ndarray):
            similarity matrix dependent on method
    '''

    methods = ['cov', 'pearson', 'R', 'cosine_similarity']
    assert np.isin(method, methods), f'RH Error: method must be one of: {methods}'

    if v2 is None:
        v2 = v1
    
    if v1.ndim == 1:
        v1 = v1[:,None]
    if v2.ndim == 1:
        v2 = v2[:,None]

    if method=='cov':
        v1_ms = v1 - np.mean(v1, axis=0) # 'mean subtracted'
        v2_ms = v2 - np.mean(v2, axis=0)
        output = (v1_ms.T @ v2_ms) / (v1.shape[0] - ddof)
    if method in ['pearson', 'R']: 
        # Below method should be as fast as numpy.corrcoef . 
        # Output should be same within precision, but 
        # numpy.corrcoef makes a doublewide matrix and can be annoying
        
        # Note: Pearson's R-value can be different than sqrt(EV) 
        # calculated below if the residuals are not orthogonal to the 
        # prediction. Best to use EV for R^2 if unsure
        v1_ms = v1 - np.mean(v1, axis=0) # 'mean subtracted'
        v2_ms = v2 - np.mean(v2, axis=0)
        output = (v1_ms.T @ v2_ms) / np.sqrt(np.sum(v1_ms**2, axis=0, keepdims=True).T * np.sum(v2_ms**2, axis=0, keepdims=True))
    if method=='cosine_similarity':    
        output = (v1 / (np.linalg.norm(v1 , axis=0, keepdims=True))).T  @ (v2  / np.linalg.norm(v2 , axis=0, keepdims=True))
    return output


def batched_covariance(X, batch_size=1000, device='cpu'):
    """
    Batched covariance matrix calculation.
    Allows for large datasets to be processed in batches on GPU.
    RH 2022

    Args:
        X (np.ndarray or torch.Tensor):
            2D array of shape (n_samples, n_features)
        batch_size (int):
            Number of samples to process at a time.
        device (str):
            Device to use for computation.
            
    Returns:
        cov (np.ndarray):
    """
    X_dl1 = list(indexing.make_batches(np.arange(X.shape[1]), batch_size=batch_size, return_idx=True))
    X_dl2 = list(indexing.make_batches(np.arange(X.shape[1]), batch_size=batch_size, return_idx=True))

    if torch.is_tensor(X):
        X_cov = torch.zeros(X.shape[1], X.shape[1], device=device)
    else:
        X_cov = np.zeros((X.shape[1], X.shape[1]))
    
    n_batches = X.shape[1] // batch_size
    for ii, (X_batch_i, idx_batch_i) in enumerate(tqdm(X_dl1, total=n_batches, leave=False, desc='outer loop')):
        for jj, (X_batch_j, idx_batch_j) in enumerate(tqdm(X_dl2, total=n_batches, leave=False, desc='inner loop')):
            x_t = X[:,idx_batch_i[0]:idx_batch_i[-1]].T
            x   = X[:,idx_batch_j[0]:idx_batch_j[-1]]
            if device != 'cpu':
                x_t = x_t.to(device)
                x = x.to(device)

            X_cov[idx_batch_i[0]:idx_batch_i[-1], idx_batch_j[0]:idx_batch_j[-1]] = x_t @ x
    return X_cov


def batched_matrix_multiply(X1, X2, batch_size1=1000, batch_size2=1000, device='cpu'):
    """
    Batched matrix multiplication of two matrices.
    Allows for multiplying huge matrices together on a GPU.
    RH 2022

    Args:
        X1 (np.ndarray or torch.Tensor):
            first matrix. shape (n_samples1, n_features1).
        X2 (np.ndarray or torch.Tensor):
            second matrix. shape (n_samples2, n_features2).
        batch_size1 (int):
            batch size for first matrix.
        batch_size2 (int):
            batch size for second matrix
        device (str):
            device to use for computation and output

    Returns:
        X1_X2 (np.ndarray or torch.Tensor):
    """
    X1_dl = indexing.make_batches(np.arange(X1.shape[1]), batch_size=batch_size1, return_idx=True)
    X2_dl = indexing.make_batches(np.arange(X2.shape[1]), batch_size=batch_size2, return_idx=True)

    if torch.is_tensor(X1):
        Y = torch.zeros(X1.shape[1], X2.shape[1], device=device)
    else:
        Y = np.zeros((X1.shape[1], X2.shape[1]))
    
    n_batches1 = X1.shape[1] // batch_size1
    n_batches2 = X2.shape[1] // batch_size2
    for ii, (X_batch_i, idx_batch_i) in enumerate(tqdm(X1_dl, total=n_batches1, leave=False, desc='outer loop')):
        for jj, (X_batch_j, idx_batch_j) in enumerate(X2_dl):
            x1_t = X1[:,idx_batch_i[0]:idx_batch_i[-1]].T
            x2   = X2[:,idx_batch_j[0]:idx_batch_j[-1]]
            if device != 'cpu':
                x1_t = x1_t.to(device)
                x2 = x2.to(device)

            Y[idx_batch_i[0]:idx_batch_i[-1], idx_batch_j[0]:idx_batch_j[-1]] = x1_t @ x2
    return Y
    # X1_dl = indexing.make_batches(X1.T, batch_size=batch_size1, return_idx=True)
    # X2_dl = indexing.make_batches(X2.T, batch_size=batch_size2, return_idx=True)

    # if torch.is_tensor(X1):
    #     Y = torch.zeros(X1.shape[1], X2.shape[1], device=device)
    # else:
    #     Y = np.zeros((X1.shape[1], X2.shape[1]))
    
    # n_batches1 = X1.shape[1] // batch_size1
    # n_batches2 = X2.shape[1] // batch_size2
    # for ii, (X_batch_i, idx_batch_i) in enumerate(tqdm(X1_dl, total=n_batches1, leave=False, desc='outer loop')):
    #     for jj, (X_batch_j, idx_batch_j) in enumerate(X2_dl):
    #         Y[idx_batch_i[0]:idx_batch_i[-1], idx_batch_j[0]:idx_batch_j[-1]] = X_batch_i @ X_batch_j.T
    # return Y


def similarity_to_distance(x, fn_toUse=1, a=1, b=0, eps=0):
    """
    Convert similarity metric to distance metric.
    RH 2022

    Args:
        x (value or array):
            similarity metric.
        fn_toUse (int from 1 to 7):
            Sets the function to use.
        a (float):
            Scaling parameter.
        b (float):
            Shifting parameter.
        eps (float):
            Small value to add to output.

    """
    if fn_toUse == 1:
        d = ((b+1) / (x**a)) -1 # fn 1: 1/x
    if fn_toUse == 2:
        d = np.exp(((-x+b)**a)) # fn 2: exp(-x)
    if fn_toUse == 3:
        d = np.arctan(a*(-x+b)) # fn 3: arctan(-x)
    if fn_toUse == 4:
        d = b - x**a # fn 4: 1-x
    if fn_toUse == 5:
        d = np.sqrt(1-(x+b))**a # fn 5: sqrt(1-x)
    if fn_toUse == 6:
        d = -np.log((x*a)+b) # fn 6: -log(x)
    
    return d + eps
    

def cp_reconstruction_EV(tensor_dense, tensor_CP):
    """
    Explained variance of a reconstructed tensor using
    by a CP tensor (similar to kruskal tensor).
    RH 2023

    Args:
        tensor_dense (np.ndarray or torch.Tensor):
            Dense tensor to be reconstructed. shape (n_samples, n_features)
        tensor_CP (tensorly CPTensor or list of np.ndarray/torch.Tensor):
            CP tensor.
            If a list of factors, then each factor should be a 2D array of shape
            (n_samples, rank).
            Can also be a tensorly CPTensor object.
    """
    tensor_rec = None
    try:
        import tensorly as tl
        if isinstance(tensor_CP, tl.cp_tensor.CPTensor):
            tensor_rec = tl.cp_to_tensor(tensor_CP)
    except ImportError as e:
        raise ImportError('tensorly not installed. Please install tensorly or provide a list of factors as the tensor_CP argument.')
    
    if tensor_rec is None:
        assert isinstance(tensor_CP, list), 'tensor_CP must be a list of factors'
        assert all([isinstance(f, (np.ndarray, torch.Tensor)) for f in tensor_CP]), 'tensor_CP must be a list of factors'
        tensor_rec = indexing.cp_to_dense(tensor_CP)

    if isinstance(tensor_dense, torch.Tensor):
        var = torch.var
    elif isinstance(tensor_dense, np.ndarray):
        var = np.var
    ev = 1 - (var(tensor_dense - tensor_rec) / var(tensor_dense))
    return ev
    

##########################################    
########### Linear Assignment ############
##########################################

def best_permutation(mat1 , mat2 , method='pearson'):
    '''
    This function compares the representations of two sets of vectors (columns
     of mat1 and columns of mat2).
    We assume that the vectors in mat1 and mat2 are similar up to a permutation.
    We therefore find the 'best' permutation that maximizes the similarity
     between the sets of vectors
    RH 2021
    
    Args:
        mat1 (np.ndarray): 
            a 2D array where the columns are vectors we wish to match with mat2
        mat2 (np.ndarray): 
            a 2D array where the columns are vectors we wish to match with mat1
        method (string)  : 
            defines method of calculating pairwise similarity between vectors:
                'pearson' or 'cosine_similarity'
        
    Returns:
        sim_avg (double): 
            the average similarity between matched vectors. Units depend on 
            method
        sim_matched (double): 
            the similarity between each pair of matched vectors.
        ind1 (int): 
            indices of vectors in mat1 matched to ind2 in mat2 (usually just 
            sequential for ind1)
        ind2 (int): 
            indices of vectors in mat2 matched to ind1 in mat1
    '''
    corr = mat1.T @ mat2
    ind1 , ind2 = scipy.optimize.linear_sum_assignment(corr, maximize=True)
    sim_matched = np.zeros(len(ind1))
    for ii in range(len(ind1)):
        if method=='pearson':
            sim_matched[ii] = np.corrcoef(mat1[:,ind1[ii]] , mat2[:,ind2[ii]])[0][1]
        if method=='cosine_similarity':
            sim_matched[ii] = pairwise_similarity( mat1[:,ind1[ii]] , mat2[:,ind2[ii]] , 'cosine_similarity')

    sim_avg = np.mean(sim_matched)
    return sim_avg , sim_matched , ind1.astype('int64') , ind2.astype('int64')

def self_similarity_pairwise(mat_set , method='pearson'):
    '''
    This function compares sets of 2-D matrices within a 3-D array using the 
    'best_permutation' function.
    We assume that the vectors within the matrices are similar up to a 
    permutation.
    We therefore find the 'best' permutation that maximizes the similarity 
    between the sets of vectors within each matrix.
    RH 2021
    
    Args:
        mat_set (np.ndarray): 
            a 3D array where the columns within the first two dims are vectors
             we wish to match with the columns from matrices from other slices
             in the third dimension
        method (string): 
            defines method of calculating pairwise similarity between vectors:
             'pearson' or 'cosine_similarity'

    Returns:
        same as 'best_permutation', but over each combo
        combos: combinations of pairwise comparisons
    '''
    
    import itertools

    n_repeats = mat_set.shape[2]
    n_components = mat_set.shape[1]

    combos = np.array(list(itertools.combinations(np.arange(n_repeats),2)))
    n_combos = len(combos)

    corr_avg = np.zeros((n_combos))
    corr_matched = np.zeros((n_components , n_combos))
    ind1 = np.zeros((n_components , n_combos), dtype='int64')
    ind2 = np.zeros((n_components , n_combos), dtype='int64')
    for i_combo , combo in enumerate(combos):
        corr_avg[i_combo] , corr_matched[:,i_combo] , ind1[:,i_combo] , ind2[:,i_combo]  =  best_permutation(mat_set[:,:,combo[0]]  ,  mat_set[:,:,combo[1]] , method)
    # print(corr_avg)
    return corr_avg, corr_matched, ind1, ind2, combos


def enumerate_paths(edges):
    """
    From Caleb Weinreb 2022

    Create a list of all paths in a directed graph
    
    Parameters
    ----------
    edges: list of tuples
        Edges in the graph as tuples (i,j) for each edge i->j. The
        edges are assumed to be in topopological order, meaning
        edge (i,j) is listed prior to edge (k,l) whenever j==k.
        
    Returns
    -------
    paths: list of tuples
        All directed paths as tuples of node indexes
    """
    child_map = {parent:[] for parent,child in edges}
    for parent,child in edges: child_map[parent].append(child)
    leaf_nodes = [child for _,child in edges if not child in child_map]
    sub_paths = {leaf:[[leaf]] for leaf in leaf_nodes}
    for parent,_ in edges[::-1]:
        if not parent in sub_paths:
            sub_paths[parent] = [[parent]]
            for child in child_map[parent]:
                for path in sub_paths[child]:
                    sub_paths[parent].append([parent]+path)
    return [tuple(p) for p in sum(sub_paths.values(),[])]
def maximum_directed_matching(weights, partition):
    """
    From Caleb Weinreb 2022

    Find the "maximum directed matching" in a weighted n-partite graph. 
    
    Let $G$ be a directed graph with nodes $\{1,...,N\}$ and weighted 
    edges $E_{ij}$. Assume that the nodes of $G$ are partitioned into 
    groups, where $P_i$ denotes the group-label for node $i$, and that 
    all the edges $E_{ij}$ satisfy $P_i < P_j$. In other words, all 
    edges go from lower to higher partition index. 
    
    We define a collection of node-sets as a "directed matching" if 
    each set forms a connected component in $G$ and no node belongs to
    more than one set. The maximum directed matching is defined as the 
    directed matching that maximizes the sum edge weights across all 
    edges that originate and terminate in the same node-set. 

    Parameters
    ----------
    weights : sparse matrix, shape=(N,N)
        The set of edge weights in the graph where N is the total 
        number of nodes. ``weights[i,j]`` is the weight of the edge 
        (i->j). A weight of 0 implies no edge. 

    partition : ndarray, shape=(N,)
        The partition label for each node in the graph. We require
        ``partition[i] < partition[j]`` whenever ``weights[i,j] != 0``
            
    Returns
    -------
    matching: list of tuples
        Maximum directed matching as a list of node-index tuples. 
    """
    import pulp
    
    # check that the edges are compatible with the partition
    sources,targets = weights.nonzero()
    assert np.all(partition[sources] < partition[targets]), 'Invalid weights/partition combination'
    
    # enumerate all possible node groupings and their score
    edge_order = np.argsort(partition[sources])
    sorted_edges = list(zip(sources[edge_order], targets[edge_order]))
    paths = enumerate_paths(sorted_edges)
    path_weights = [sum([weights[i,j] for i in p for j in p]) for p in paths]
    
    # configure an ILP solver with pulp
    problem = pulp.LpProblem('matching', pulp.LpMaximize)
    
    # initialize variables
    path_vars = pulp.LpVariable.dicts('paths', paths, cat=pulp.LpBinary)
    problem += pulp.lpSum([w*path_vars[p] for p,w in zip(paths, path_weights)])

    # set constraints
    path_membership = {n:[] for n in range(weights.shape[0])}
    for path in paths:
        for n in path: path_membership[n].append(path)
    for n in path_membership: 
        problem += pulp.lpSum(path_vars[path] for path in path_membership[n]) <= 1
        
    # solve and extract results
    problem.solve()
    matching = [p for p,v in path_vars.items() if v.value()==1]
    return matching



import numpy as np
import scipy.stats

def ttest_paired_ratio(a, b):
    """
    Computes the ratio paired t-test between two arrays of data.
    This is useful for comparing ratios of data, and can be
     thought of as a geometric mean paired t-test.
    The ratio t-test takes the logarithm of the ratio of
     a vs. b and then tests the null hypothesis that the
     mean is zero.
    
    RH 2023

    Args:
        a (np.ndarray or torch.Tensor):
            Array of data. Paired 1-to-1 with b.
        b (np.ndarray or torch.Tensor):
            Array of data. Paired 1-to-1 with a.
    """
    # Compute ratio
    ratio = a/b
    # Compute log of ratio
    log_ratio = np.log(ratio)
    # Compute mean of log ratio
    mean_log_ratio = np.mean(log_ratio)
    # Compute standard error of log ratio
    stderr_log_ratio = scipy.stats.sem(log_ratio)
    # Compute t-statistic
    t_stat = mean_log_ratio/stderr_log_ratio
    # Compute p-value
    p_val = scipy.stats.t.sf(np.abs(t_stat), len(a)-1)*2

    ## above is equivalent to:
    # t_stat, p_val = scipy.stats.ttest_rel(np.log(a/b), np.zeros_like(a))
    return p_val


def geometric_mean(a):
    """
    Computes the geometric mean of an array of data.
    This is useful for computing the geometric mean of ratios.
    
    RH 2023

    Args:
        a (np.ndarray or torch.Tensor):
            Array of data.
    """
    return np.exp(np.mean(np.log(a)))


def sparsity(a, axis=0):
    """
    A normalized dispersion index. Varies from 0 (dense) to 1 (sparse).
    Computes the lifetime sparseness of an array of data.
    From Vinje & Gallant 2000; and Willmore, Mazer, & Gallant 2011.
    
    RH 2023

    Args:
        a (np.ndarray or torch.Tensor):
            Array of data.
        axis (int):
            Axis along which to compute sparsity.

    Returns:
        float: Sparsity value. 0 (dense) to 1 (sparse).
    """
    return 1 - ((a.mean(axis) ** 2) / (a ** 2).mean(axis))


def error_interval(
    data, 
    axis=0,
    error="sem",
    confidence=0.95, 
):
    """
    Calculate the mean and the confidence interval of the mean.

    Args:
    data (np.ndarray): 
        The input data. Can be a list or a numpy array.
    axis (int): 
        Axis along which to compute the mean and confidence interval. Default is 0.
    error (str):
        Type of error to compute. \n
            * "std" - Standard deviation.
            * "sem" - Standard error of the mean.
            * "ci" - Confidence interval. \n
    confidence (float): 
        The confidence level for the interval. Only used if error is "ci".

    Returns:
    tuple:
        mean (float): 
            The mean of the data along the specified axis.
        lower (float):
            The lower bound of the confidence interval.
        upper (float):
            The upper bound of the confidence interval.
    """

    assert confidence > 0 and confidence < 1, "Confidence level must be between 0 and 1."
    assert isinstance(data, np.ndarray), "Data must be a numpy array."

    # Calculate the size along the specified axis
    n = data.shape[axis]

    # Compute the mean along the specified axis

    if error in ["std", "sem", "ci"]:
        m = np.mean(data, axis=axis)
    
        # Compute the standard deviation along the specified axis
        sd = np.std(data, axis=axis)
        if error in ["sem", "ci"]:
            # Compute the standard error of the mean along the specified axis
            se = sd / np.sqrt(n)
        if error == "ci":
            # Compute the margin of error
            h = se * scipy.stats.t.ppf((1 + confidence) / 2., n-1)
            mean, lower, upper = m, m-h, m+h
        if error == "sem":
            mean, lower, upper = m, m-se, m+se
        if error == "std":
            mean, lower, upper = m, m-sd, m+sd
    else:
        raise ValueError("Error must be one of 'std', 'sem', or 'ci'.")
    
    return mean, lower, upper


import sklearn.decomposition
import numpy as np
import scipy.interpolate
import matplotlib.pyplot as plt

import torch
from torch.utils.data import Dataset, DataLoader

import copy

# import cuml
# import cuml.decomposition
# import cupy

import gc

from tqdm.notebook import tqdm

###########################
########## PCA ############
###########################

def simple_pca(X , n_components=None , mean_sub=True, zscore=False, plot_pref=False , n_PCs_toPlot=2):
    """
    Performs PCA on X.
    RH 2021

    Args:
        X (np.ndarray):
            Data to be decomposed.
            2-D array. Columns are features, rows are samples.
        n_components (int):
            Number of components to keep. If None, then
             n_components = X.shape[1]
        mean_sub (bool):
            Whether or not to mean subtract ('center') the
             columns.
        zscore (bool):
            Whether or not to z-score the columns. This is
             equivalent to doing PCA on the correlation-matrix.
        plot_pref (bool):
            Whether or not to plot the first n_PCs_toPlot of the
             PCA.
        n_PCs_toPlot (int):
            Number of PCs to plot.

    Returns:
        components (np.ndarray):
            The components of the decomposition.
            2-D array.
            Each column is a component vector. Each row is a
             feature weight.
        scores (np.ndarray):
            The scores of the decomposition.
            2-D array.
            Each column is a score vector. Each row is a
             sample weight.
        EVR (np.ndarray):
            The explained variance ratio of each component.
            1-D array.
            Each element is the explained variance ratio of
             the corresponding component.
    """
    if mean_sub and not zscore:
        X = X - np.mean(X, axis=0)
    if zscore:
        # X = scipy.stats.zscore(X, axis=0)
        X = X - np.mean(X, axis=0)
        stds = np.std(X, axis=0)
        X = X / stds[None,:]
    
    if n_components is None:
        n_components = X.shape[1]
    decomp = sklearn.decomposition.PCA(n_components=n_components)
    decomp.fit_transform(X)
    components = decomp.components_
    scores = decomp.transform(X)
    
    if plot_pref:
        fig , axs = plt.subplots(4 , figsize=(7,15))
        axs[0].plot(np.arange(n_components)+1,
                    decomp.explained_variance_ratio_)
        axs[0].set_xscale('log')
        axs[0].set_xlabel('component #')
        axs[0].set_ylabel('explained variance ratio')

        axs[1].plot(np.arange(n_components)+1, 
                    np.cumsum(decomp.explained_variance_ratio_))
        axs[1].set_xscale('log')
        axs[1].set_ylabel('cumulative explained variance ratio')

        axs[2].plot(scores[:,:n_PCs_toPlot])
        axs[2].set_xlabel('sample num')
        axs[2].set_ylabel('a.u.')

        axs[3].plot(components.T[:,:n_PCs_toPlot])
        axs[3].set_xlabel('feature num')
        axs[3].set_ylabel('score')
    
    return components , scores , decomp.explained_variance_ratio_


def torch_pca(
    X_in, 
    device='cpu', 
    mean_sub=True, 
    zscore=False, 
    rank=None, 
    return_cpu=True, 
    return_numpy=False,
    cuda_empty_cache=True,
):
    """
    Principal Components Analysis for PyTorch.
    If using GPU, then call torch.cuda.empty_cache() after.
    RH 2021

    Args:
        X_in (torch.Tensor or np.ndarray):
            Data to be decomposed.
            2-D array. Columns are features, rows are samples.
            PCA will be performed column-wise.
        device (str):
            Device to use. ie 'cuda' or 'cpu'. Use a function 
             torch_helpers.set_device() to get.
        mean_sub (bool):
            Whether or not to mean subtract ('center') the 
             columns.
        zscore (bool):
            Whether or not to z-score the columns. This is 
             equivalent to doing PCA on the correlation-matrix.
        rank (int):
            Maximum estimated rank of decomposition. If None,
             then rank is X.shape[1]
        return_cpu (bool):  
            Whether or not to force returns/outputs to be on 
             the 'cpu' device. If False, and device!='cpu',
             then returns will be on device.
        return_numpy (bool):
            Whether or not to force returns/outputs to be
             numpy.ndarray type.
        cuda_empty_cache (bool):
            Whether or not to call torch.cuda.empty_cache()
            Only relevant if device!='cpu'.

    Returns:
        components (torch.Tensor or np.ndarray):
            The components of the decomposition. 
            2-D array.
            Each column is a component vector. Each row is a 
             feature weight.
        scores (torch.Tensor or np.ndarray):
            The scores of the decomposition.
            2-D array.
            Each column is a score vector. Each row is a 
             sample weight.
        singVals (torch.Tensor or np.ndarray):
            The singular values of the decomposition.
            1-D array.
            Each element is a singular value.
        EVR (torch.Tensor or np.ndarray):
            The explained variance ratio of each component.
            1-D array.
            Each element is the explained variance ratio of
             the corresponding component.
    """
    
    if isinstance(X_in, torch.Tensor) == False:
        X = torch.from_numpy(X_in).to(device)
    elif str(X_in.device) != device:
        X = X_in.to(device)
    else:
        X = X_in
        
    if mean_sub and not zscore:
        X = X - torch.mean(X, dim=0)
    if zscore:
        X = X - torch.mean(X, dim=0)
        stds = torch.std(X, dim=0)
        idx_zeroStd = torch.where(stds==0)[0]
        stds[idx_zeroStd] = 1.0
        X = X / stds
        
    if rank is None:
        rank = min(list(X.shape))
    
    (U,S,V) = torch.pca_lowrank(X, q=rank, center=False, niter=2)
    components = V
    scores = torch.matmul(X, V[:, :rank])

    singVals = (S**2)/(len(S)-1)
    EVR = (singVals) / torch.sum(singVals)
    
    if return_cpu:
        components = components.cpu()
        scores = scores.cpu()
        singVals = singVals.cpu()
        EVR = EVR.cpu()
    if return_numpy:
        components = components.cpu().numpy()
        scores = scores.cpu().numpy()
        singVals = singVals.cpu().numpy()
        EVR = EVR.cpu().numpy()

    if cuda_empty_cache and device!='cpu':
        torch.cuda.empty_cache()
        gc.collect()
        
    return components, scores, singVals, EVR

def unmix_pcs(pca_components, weight_vecs):
    """
    Transforms weight_vecs into pca_components space
    RH 2021
    """
    if weight_vecs.ndim == 1:
        weight_vecs = weight_vecs[:,None]
    if type(pca_components) == np.ndarray:
        zeros = np.zeros
    elif type(pca_components) == torch.Tensor:
        zeros = torch.zeros
    
    mixing_vecs = zeros((pca_components.shape[1], weight_vecs.shape[1]))
    mixing_vecs[:weight_vecs.shape[0],:] = weight_vecs

    return pca_components @ mixing_vecs


def dimensionality_pca(
    X, 
    ev=[0.10,0.25,0.50,0.90,0.95], 
    mean_sub=True, 
    zscore=False, 
    device='cpu', 
    kwargs_torchPCA={}
):
    """
    Returns the number of components needed to explain a certain 
     amount of variance. 
    Calculates the 'linear embedding dimensionality'.
    Uses torch_pca() function to perform PCA.
    RH 2022

    Args:
        X (np.ndarray or torch.Tensor):
            Data to be decomposed.
            2-D array. Columns are features, rows are samples.
        ev (list or np.ndarray):
            List of explained variance ratios to calculate the
             number of components needed to explain.
            ie [0.10,0.25,0.50,0.90,0.95]
        mean_sub (bool):
            Whether or not to mean subtract ('center') the
             columns.
        zscore (bool):
            Whether or not to z-score the columns. This is
             equivalent to doing PCA on the correlation-matrix.
        device (str):
            Device to use. ie 'cuda' or 'cpu'. Use a function
             torch_helpers.set_device() to get.

    Returns:
        n_components (np.ndarray):
            Number of components needed to explain the
             corresponding ev.
            Same array type (torch.Tensor or np.array) as X.
    """
    ## make sure ev is a list or np.ndarray
    ev = [ev] if isinstance(ev, (int,float)) else ev
    assert isinstance(ev, (list,np.ndarray)), 'RH ERROR: ev must be a list or np.ndarray.'
    ev = np.array(ev)

    ## assert there are no nans
    assert np.any(np.isnan(X)) == False, 'RH ERROR: X must not contain NaNs.'

    comps, scores, sv, EVR = torch_pca(
        X_in=X, 
        device=device,
        return_numpy=True,
        mean_sub=mean_sub,
        zscore=zscore,
        **kwargs_torchPCA,
    )
    EVR_cumsum = np.cumsum(np.concatenate([[0], EVR]))
    interp = scipy.interpolate.interp1d(x=EVR_cumsum, y=np.arange(0, len(EVR_cumsum)), kind='linear')
    return interp(ev)


#######################################
########## Incremental PCA ############
#######################################

class IPCA_Dataset(Dataset):
    """
    see incremental_pca for demo
    """
    def __init__(self, 
                 X, 
                 mean_sub=True,
                 zscore=False,
                 preprocess_sample_method='random',
                 preprocess_sample_num=100,
                 device='cpu',
                 dtype=torch.float32):
        """
        Make a basic dataset.
        RH 2021

        Args:
            X (torch.Tensor or np.array):
                Data to make dataset from.
                2-D array. Columns are features, rows are samples.
            mean_sub (bool):
                Whether or not to mean subtract ('center') the
                 columns.
            zscore (bool):
                Whether or not to z-score the columns. This is
                 equivalent to doing PCA on the correlation-matrix.
            preprocess_sample_method (str):
                Method to use for sampling for mean_sub and zscore.
                'random' - uses random samples (rows) from X.
                'first' - uses the first rows of X.
            preprocess_sample_num (int):
                Number of samples to use for mean_sub and zscore.
            device (str):
                Device to use.
            dtype (torch.dtype):
                Data type to use.
        """
        # Upgrade here for using an on the fly dataset.
        self.X = torch.as_tensor(X, dtype=dtype, device=device) # first (0th) dim will be subsampled from
        
        self.n_samples = self.X.shape[0]

        self.mean_sub = mean_sub
        self.zscore = zscore

        if mean_sub or zscore:
            if preprocess_sample_method == 'random':
                self.preprocess_inds = torch.randperm(preprocess_sample_num)[:preprocess_sample_num]
            elif preprocess_sample_method == 'first':
                self.preprocess_inds = torch.arange(preprocess_sample_num)
            else:
                raise ValueError('preprocess_sample_method must be "random" or "first"')
            self.preprocess_inds = self.preprocess_inds.to(device)

        # Upgrade here for using an on the fly dataset.
        if mean_sub:
            self.mean_vals = torch.mean(self.X[self.preprocess_inds,:], dim=0)
        if zscore:
            self.std_vals = torch.std(self.X[self.preprocess_inds,:], dim=0)
        
    def __len__(self):
        return self.n_samples
    
    def __getitem__(self, idx):
        """
        Returns a single sample.

        Args:
            idx (int):
                Index of sample to return.
        """
        if self.mean_sub or self.zscore:
            out = self.X[idx,:] - self.mean_vals
        else:
            out = self.X[idx,:]

        if self.zscore:
            out = out / self.std_vals
            
        return out, idx

def incremental_pca(dataloader,
                    method='sklearn',
                    method_kwargs=None,
                    return_cpu=True,
                    ):
    """
    Incremental PCA using either sklearn or cuml.
    Keep batches small-ish to remain performat (~64 samples).
    RH 2021

    Args:
        dataloader (torch.utils.data.DataLoader):
            Data to be decomposed.
        method (str):
            Method to use.
            'sklearn' : sklearn.decomposition.PCA
            'cuml' : cuml.decomposition.IncrementalPCA
        method_kwargs (dict):
            Keyword arguments to pass to method.
            See method documentation for details.
        device (str):
            Device to use.
            Only used if method is 'cuml'
        return_cpu (bool):  
            Whether or not to force returns/outputs to be on 
             the 'cpu' device. If False, and device!='cpu',
             then returns will be on device.
        return_numpy (bool):
            Whether or not to force returns/outputs to be
             numpy.ndarray type.

    Returns:
        components (torch.Tensor or np.ndarray):
            The components of the decomposition. 
            2-D array.
            Each column is a component vector. Each row is a 
             feature weight.
        EVR (torch.Tensor or np.ndarray):
            The explained variance ratio of each component.
            1-D array.
            Each element is the explained variance ratio of
             the corresponding component.
        object_params (dict):
            Dictionary of parameters used to create the
             decomposition.

    demo:
    import torch_helpers, cupy_helpers

    cupy_helpers.set_device() # calls: cupy.cuda.Device(DEVICE_NUM).use()

    dataset = decomposition.IPCA_Dataset(   X, 
                                            mean_sub=True,
                                            zscore=False,
                                            preprocess_sample_method='random',
                                            preprocess_sample_num=10000,
                                            device='cpu',
                                            dtype=torch.float32)
    dataloader = torch.utils.data.DataLoader(   dataset, 
                                                batch_size=64, 
                                                drop_last=True,
                                                shuffle=False, 
                                                num_workers=0, 
                                                pin_memory=False)

    cuml_kwargs = {
                "handle": cuml.Handle(),
                "n_components": 20,
                "whiten": False,
                "copy": False,
                "batch_size": None,
                "verbose": True,
                "output_type": None
    }

    sk_kwargs = {
                    "n_components": 20,
                    "whiten": False,
                    "copy": False,
                    "batch_size": None,
    }

    components, EVR, ipca = decomposition.incremental_pca(dataloader,
                                    method='cuml',
                                    method_kwargs=cuml_kwargs,
                                    return_cpu=True)
    scores = decomposition.ipca_transform(dataloader, components)
    """
    
    if method_kwargs is None:
        method_kwargs = {}

    if method == 'sklearn':
        ipca = sklearn.decomposition.IncrementalPCA(**method_kwargs)
    elif method == 'cuml':
        ipca =    cuml.decomposition.IncrementalPCA(**method_kwargs)
    
    for iter_batch, batch in enumerate(tqdm(dataloader)):
        if method == 'sklearn':
            batch = batch[0].cpu().numpy()
        if method == 'cuml':
            batch = cupy.asarray(batch[0])
        ipca.partial_fit(batch)
    
    if (return_cpu) and (method=='cuml'):
        components = ipca.components_.get()
    else:
        components = ipca.components_

    EVR = ipca.explained_variance_ratio_

    return components, EVR, ipca

def ipca_transform(dataloader, components):
    """
    Transform data using incremental PCA.
    RH 2020

    Args:
        dataloader (torch.utils.data.DataLoader):
            Data to be decomposed.
        components (torch.Tensor or np.ndarray):
            The components of the decomposition. 
            2-D array.
            Each column is a component vector. Each row is a 
             feature weight.
    """
    out = []
    for iter_batch, batch in enumerate(dataloader):
        out.append(batch[0] @ components.T)
    return torch.cat(out, dim=0)